#
# HRG.py
#
# J. Goswami, D. Clarke, Sipaz Sharma
#
# Collection of methods pertaining to hadron resonance gas calculations. Based in part 
# on code by F. Karsch.
#

import numpy as np
import sympy as sy
import warnings
from scipy.special import lambertw, kn
from sympy import Sum, symbols, Indexed, lambdify, LambertW, exp
import latqcdtools.base.logger as logger
from latqcdtools.math.num_int import integrateFunction
from latqcdtools.base.check import ignoreUnderflow
from latqcdtools.base.utilities import envector, unvector


ignoreUnderflow()
warnings.filterwarnings("error")
warnings.filterwarnings("ignore", category=DeprecationWarning)


APPROX_KAON_MASS = 500   # Mass cutoff in [MeV] for Boltzmann approximation.


def RMS_mass(Nt, T):
    a = 6924.46
    b = 31881.4
    c = -1.02357e+06
    mkaon = 493.68
    mpion = 140.0
    x = 1.0/Nt/T*197.3
    pion_rms_mass = mpion + a*x**2+b*x**4+c*x**6
    del2 = pion_rms_mass**2 - mpion**2
    kaon_rms_mass = np.sqrt(mkaon**2+del2)
    return pion_rms_mass, kaon_rms_mass


def LCP_init_NS0(muB):
    """ 
    Give a good initial guess for NS=0 LCP. 
    """
    dS = 0.214
    eS = 0.161
    dQ = 0.0211
    eQ = 0.106
    muQ = -dQ / (1.0 + eQ * muB)
    muS = dS / (1.0 + eS * muB)
    return muQ, muS


def dmuh(order, muh):
    """ 
    d^order/dmuh^order derivative of muh 
    """
    if order == 0:
        return muh
    elif order == 1:
        return 1
    else:
        return 0


class HRGbase:

    """ 
    Hadron resonance gas base class. Here we collect methods and attributes that all HRG-type classes should have
    in common. Mass=mass of the Hadron/resonance , g=spin degenerecy , w= fermi(-1)/bose(1) statistics.
    B, Q, S, and C are respectively the baryon number, electric charge, strangeness, and charm of each state. 
    """

    def __init__(self, Mass, g, w, B, S, Q, C=None):
        self.Mass = Mass
        self.g = g
        self.w = w
        self.B = B
        self.Q = Q
        self.S = S
        if C is None:
            self.C = np.zeros(len(Mass))
        else:
            self.C = C

    def __repr__(self) -> str:
        return "HRGbase"

    def muN_div_T(self, k, muB_div_T, muQ_div_T, muS_div_T, muC_div_T):
        """ 
        mu_X * N_X, X = (B,Q,S,C) 
        """
        return self.B[k]*muB_div_T + self.Q[k]*muQ_div_T + self.S[k]*muS_div_T + self.C[k]*muC_div_T

    def z(self, k, muB_div_T, muQ_div_T, muS_div_T, muC_div_T):
        """ 
        e^(mu_X*N_X/T) , X = (B,Q,S,C) 
        """
        return np.exp(self.muN_div_T(k, muB_div_T, muQ_div_T, muS_div_T, muC_div_T))


class HRG(HRGbase):
    """ 
    HRG implemented through Taylor expasion of logarithm. For more information please see, e.g.
    Physics Letters B 695 (2011) 136–142 or especially arXiv:2011.02812.
    You can optionally adjust NMAX_light and NMAX_heavy, which control the number of terms to keep in the
    Taylor expansion for species that are respectively lighter and heavier than the Kaon.
    Our pressure is given in terms of a Taylor series involving modified Bessel functions of the second kind, which
    needs to be truncated at some order. These functions get strongly suppressed when their argument is large.
    In our case, this argument is proportional to the mass. Hence we will sometimes take the Boltzmann
    approximation, i.e. that the mass is large compared to the temperature. In this limit, even fewer terms of the
    expansion need to be kept. Doing so boosts performance.

    We work in the grand canonical ensemble, i.e. we consider P(V,T,mu_i/T). Hence in this class, derivatives w.r.t.
    one of those N_chemical_potentials + 2 variables assume all others are held fixed. 
    """

    def __init__(self, Mass, g, w, B, S, Q, C=None, NMAX_light=21, NMAX_heavy=2):
        HRGbase.__init__(self, Mass, g, w, B, S, Q, C)
        self.NMAX_light = NMAX_light
        self.NMAX_heavy = NMAX_heavy

    def __repr__(self) -> str:
        return "HRG"

    def Nmax(self, k):
        """ Keep Nmax terms of the Taylor expansion. """
        if self.Mass[k] < APPROX_KAON_MASS:
            return self.NMAX_light
        else:
            return self.NMAX_heavy

    # n represents the nth order of a Taylor expansion of a logarithm.
    # k represents the kth state that appears in the table of resonances.
    # n=1 is Boltzmann approximation

    def factor(self, k, n, T):
        """ 
        (m/T)^2 g eta^(n+1) / 2pi^2 n^2 
        """
        return (self.Mass[k]/T)**2 * self.g[k] * self.w[k]**(n+1) / (2*np.pi**2*n**2)

    def P_div_T4(self, T, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        P = 0.
        for k in range(len(self.Mass)):
            for n in range(1, self.Nmax(k)):
                P += self.factor(k, n, T) \
                        * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                        * kn(2, (n * self.Mass[k] / T))
        return P

    def E_div_T4(self, T, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        eps = 0.
        for k in range(len(self.Mass)):
            for n in range(1, self.Nmax(k)):
                x = self.Mass[k]*n/T
                eps += self.factor(k, n, T) \
                        * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                        * (kn(2, x) * 3 + kn(1, x) * x)
        return eps

    def S_div_T3(self, T, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        """ 
        s = e + p - mu_i n_i 
        """
        NB = self.gen_chi(T, B_order=1, S_order=0, Q_order=0, C_order=0, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        NS = self.gen_chi(T, B_order=0, S_order=1, Q_order=0, C_order=0, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        NQ = self.gen_chi(T, B_order=0, S_order=0, Q_order=1, C_order=0, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        NC = self.gen_chi(T, B_order=0, S_order=0, Q_order=0, C_order=1, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        muxN_div_T = NB*muB_div_T + NS*muS_div_T + NQ*muQ_div_T + NC*muC_div_T
        return self.E_div_T4(T, muB_div_T=muB_div_T, muS_div_T=muS_div_T, muQ_div_T=muQ_div_T, muC_div_T=muC_div_T) \
                + self.P_div_T4(T, muB_div_T=muB_div_T, muS_div_T=muS_div_T,muQ_div_T=muQ_div_T, muC_div_T=muC_div_T) - muxN_div_T

    def ddT_E_div_T4(self, T, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        """ 
        d(E/T^4)/dT at fixed mu/T 
        """
        eps = 0.
        for k in range(len(self.Mass)):
            for n in range(1, self.Nmax(k)):
                m = self.Mass[k]
                x = m*n/T
                eps += self.factor(k, n, T)*m*n \
                        * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                        * (kn(0, x)*n*m + kn(1, x)*T)/T**3
        return eps

    def ddT_P_div_T4(self, T, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        """ 
        d(P/T^4)/dT at fixed mu/T 
        """
        P = 0.
        for k in range(len(self.Mass)):
            for n in range(1, self.Nmax(k)):
                m = self.Mass[k]
                x = m*n/T
                P += self.factor(k, n, T) \
                        * self.z(k, muB_div_T=muB_div_T,muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                        * m*n*kn(1, x) / T**2
        return P

    def CV_div_T3_mu0(self, T):
        """ 
        C_V/T^3 at mu=0. 
        """
        return 4*self.E_div_T4(T, 0, 0, 0, 0) + T*self.ddT_E_div_T4(T, 0, 0, 0, 0)

    def ddT_S_div_T3(self, T, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        """ 
        d(s/T^3)/dT at fixed mu/T 
        """
        ddTNB = self.ddT_gen_chi(T, B_order=1, S_order=0, Q_order=0, C_order=0,
                                 muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        ddTNS = self.ddT_gen_chi(T, B_order=0, S_order=1, Q_order=0, C_order=0,
                                 muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        ddTNQ = self.ddT_gen_chi(T, B_order=0, S_order=0, Q_order=1, C_order=0,
                                 muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        ddTNC = self.ddT_gen_chi(T, B_order=0, S_order=0, Q_order=0, C_order=1,
                                 muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        muxddTN_div_T = ddTNB*muB_div_T + ddTNS*muS_div_T + ddTNQ*muQ_div_T + ddTNC*muC_div_T
        return self.ddT_E_div_T4(T, muB_div_T=muB_div_T, muS_div_T=muS_div_T, muQ_div_T=muQ_div_T, muC_div_T=muC_div_T) \
                + self.ddT_P_div_T4(T, muB_div_T=muB_div_T, muS_div_T=muS_div_T, muQ_div_T=muQ_div_T, muC_div_T=muC_div_T) - muxddTN_div_T

    def gen_chi(self, T, B_order=0, S_order=0, Q_order=0, C_order=0, muB_div_T=0., muQ_div_T=0., muS_div_T=0., muC_div_T=0.):
        """ 
        chi_BQSC 
        """
        chi = 0.0
        for k in range(len(self.Mass)):
            for n in range(1, self.Nmax(k)):
                zn_Kn = self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n * kn(2, (n*self.Mass[k]/T))
                chi += (self.B[k]*n)**B_order * (self.S[k]*n)**S_order * (self.Q[k]*n)**Q_order * (self.C[k]*n)**C_order * self.factor(k, n, T) * zn_Kn
        return chi

    def ddT_gen_chi(self, T, B_order=0, S_order=0, Q_order=0, C_order=0, muB_div_T=0., muQ_div_T=0., muS_div_T=0., muC_div_T=0.):
        """ 
        d(chi_BQSC)/dT at fixed mu/T 
        """
        chi = 0.0
        for k in range(len(self.Mass)):
            for n in range(1, self.Nmax(k)):
                m = self.Mass[k]
                x = m*n/T
                chi += (self.B[k]*n)**B_order * (self.S[k]*n)**S_order * (self.Q[k]*n)**Q_order * (self.C[k]*n)**C_order \
                                              * self.factor(k, n, T) \
                                              * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                                              * m*n*kn(1, x) / T**2
        return chi

    def d2dT2_gen_chi(self, T, B_order=0, S_order=0, Q_order=0, C_order=0, muB_div_T=0., muQ_div_T=0., muS_div_T=0., muC_div_T=0.):
        """ 
        d^2(chi_BQSC)/dT^2 at fixed mu/T 
        """
        chi = 0.0
        for k in range(len(self.Mass)):
            for n in range(1, self.Nmax(k)):
                m = self.Mass[k]
                x = m*n/T
                chi += (self.B[k]*n)**B_order * (self.S[k]*n)**S_order * (self.Q[k]*n)**Q_order * (self.C[k]*n)**C_order \
                                              * self.factor(k, n, T) * m*n \
                                              * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                                              * (m*n*kn(0, x)/T**4 - 3*kn(1, x)/T**3)
        return chi

    def gen_ddmuh_E_div_T4(self, T, B_order=0, S_order=0, Q_order=0, C_order=0, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        """ 
        Arbitrary mu/T derivatives of E/T^4 at fixed T 
        """
        eps = 0.
        for k in range(len(self.Mass)):
            for n in range(1, self.Nmax(k)):
                x = self.Mass[k]*n/T
                eps += (self.B[k]*n)**B_order * (self.S[k]*n)**S_order * (self.Q[k]*n)**Q_order * (self.C[k]*n)**C_order \
                                              * self.factor(k, n, T) \
                                              * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                                              * (kn(2, x)*3 + kn(1, x)*x)
        return eps

    def gen_ddmuh_P_div_T4(self, T, B_order=0, S_order=0, Q_order=0, C_order=0, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        """ 
        Arbitrary mu/T derivatives of P/T^4 at fixed T 
        """
        P = 0.
        for k in range(len(self.Mass)):
            for n in range(1, self.Nmax(k)):
                P += (self.B[k]*n)**B_order * (self.S[k]*n)**S_order * (self.Q[k]*n)**Q_order * (self.C[k]*n)**C_order \
                                            * self.factor(k, n, T) \
                                            * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                                            * kn(2, (n*self.Mass[k]/T))
        return P

    def gen_ddmuh_S_div_T3(self, T, B_order=0, S_order=0, Q_order=0, C_order=0, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        """ 
        s = e + p - mu_i n_i 
        """
        NB = self.gen_chi(T, B_order=1, S_order=0, Q_order=0, C_order=0, muB_div_T=muB_div_T,
                          muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        NS = self.gen_chi(T, B_order=0, S_order=1, Q_order=0, C_order=0, muB_div_T=muB_div_T,
                          muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        NQ = self.gen_chi(T, B_order=0, S_order=0, Q_order=1, C_order=0, muB_div_T=muB_div_T,
                          muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        NC = self.gen_chi(T, B_order=0, S_order=0, Q_order=0, C_order=1, muB_div_T=muB_div_T,
                          muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        dmuhNB = self.gen_chi(T, B_order=1+B_order, S_order=S_order, Q_order=Q_order, C_order=C_order,
                              muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        dmuhNS = self.gen_chi(T, B_order=B_order, S_order=1+S_order, Q_order=Q_order, C_order=C_order,
                              muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        dmuhNQ = self.gen_chi(T, B_order=B_order, S_order=S_order, Q_order=1+Q_order, C_order=C_order,
                              muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        dmuhNC = self.gen_chi(T, B_order=B_order, S_order=S_order, Q_order=Q_order, C_order=1+C_order,
                              muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)
        return self.gen_ddmuh_E_div_T4(T, B_order=B_order, S_order=S_order, Q_order=Q_order, C_order=C_order, 
                                       muB_div_T=muB_div_T, muS_div_T=muS_div_T, muQ_div_T=muQ_div_T, muC_div_T=muC_div_T) \
                + self.gen_ddmuh_P_div_T4(T, B_order=B_order, S_order=S_order, Q_order=Q_order, C_order=C_order, 
                                          muB_div_T=muB_div_T, muS_div_T=muS_div_T, muQ_div_T=muQ_div_T, muC_div_T=muC_div_T) \
                - muB_div_T*dmuhNB - muQ_div_T*dmuhNQ - muS_div_T*dmuhNS - muC_div_T*dmuhNC \
                - NB*dmuh(B_order, muB_div_T) - NQ*dmuh(Q_order, muQ_div_T) - NS*dmuh(S_order, muS_div_T) - NC*dmuh(C_order, muC_div_T)

    def gen_chi_RMS(self, T, Nt, B_order=0, S_order=0, Q_order=0, C_order=0, muB_div_T=0., muQ_div_T=0., muS_div_T=0., muC_div_T=0.):
        # rms_mass[0] is for pions and rms_mass[1] is for kaons
        rms_mass = RMS_mass(Nt, T)
        chi = 0.0
        for k in range(len(self.Mass)):
            if 140 >= self.Mass[k] >= 130:
                for n in range(1, 20):
                    chi += (self.B[k]*n)**B_order * (self.S[k]*n)**S_order \
                                                  * (self.Q[k]*n)**Q_order \
                                                  * (self.C[k]*n)**C_order \
                                                  * self.w[k]**(n+1) * self.g[k] * (rms_mass[0]/T)**2 \
                                                  * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                                                  * kn(2, (n*rms_mass[0]/T)) / (np.pi*n)**2 / 2
            elif 500 >= self.Mass[k] >= 490:
                for n in range(1, 10):
                    chi += (self.B[k]*n)**B_order * (self.S[k]*n)**S_order \
                                                  * (self.Q[k]*n)**Q_order \
                                                  * (self.C[k]*n)**C_order \
                                                  * self.w[k]**(n+1) * self.g[k] * (rms_mass[1]/T)**2 \
                                                  * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                                                  * kn(2, (n*rms_mass[1]/T)) / (np.pi*n)**2 / 2
            else:
                for n in range(1, 2):
                    chi += (self.B[k]*n)**B_order * (self.S[k]*n)**S_order \
                                                  * (self.Q[k]*n)**Q_order \
                                                  * (self.C[k]*n)**C_order \
                                                  * self.factor(k, n, T) \
                                                  * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                                                  * kn(2, (n*self.Mass[k]/T))
        return chi

    def genChiFlavor(self, T, u_order=0, d_order=0, s_order=0, c_order=0, muB_div_T=0., muQ_div_T=0., muS_div_T=0., muC_div_T=0.):
        """ 
        chi_udsc 
        """
        logger.warn('HRGFlavor basis is not yet tested for finite mu...')
        chi = 0.0
        for k in range(len(self.Mass)):
            for n in range(1, self.Nmax(k)):
                chi += ((self.B[k]+self.Q[k]-self.C[k])*n)**u_order \
                    * ((2.0*self.B[k]-self.Q[k]+self.S[k])*n)**d_order \
                    * (-self.S[k]*n)**s_order \
                    * (self.C[k]*n)**c_order \
                    * self.factor(k, n, T) \
                    * self.z(k, muB_div_T=muB_div_T, muQ_div_T=muQ_div_T, muS_div_T=muS_div_T, muC_div_T=muC_div_T)**n \
                    * kn(2, (n*self.Mass[k]/T))
        return chi
# TODO: gen_chi can be implemented using multi-dimensional Leibniz rule


class HRGexact(HRGbase):
# TODO: maybe just delete this eventually, it seems to be crap

    """ 
    HRG implemented through numerical integration. 
    """

    def __init__(self, Mass, g, w, B, S, Q, C=None):
        logger.warn('ExactHRG numerical integration is not yet reliable...')
        HRGbase.__init__(self, Mass, g, w, B, S, Q, C)

    def __repr__(self) -> str:
        return "HRGexact"

    def P_div_T4(self, T, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T = envector(T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T)

        def int_wrapper(Tvec, muBvec, muSvec, muQvec, muCvec):
            P = 0.
            for k in range(len(self.Mass)):
                wz = self.w[k] * self.z(k, muB_div_T=muBvec,muQ_div_T=muQvec, muS_div_T=muSvec, muC_div_T=muCvec)
                def integrand(E):
                    exp_E_div_T = np.exp(-E/Tvec)
                    return -(wz/(3*Tvec)) * (E**2-self.Mass[k]**2)**(3/2) * exp_E_div_T / (1-wz*exp_E_div_T)
                P -= self.w[k] * self.g[k] * integrateFunction(integrand, self.Mass[k], np.inf, method='quad') / (2*np.pi**2 * Tvec**3)
            return P
        int_vec = np.vectorize(int_wrapper)
        return unvector(np.asarray(int_vec(T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T)))

    def E_div_T4(self, T, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T = envector(T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T)

        def int_wrapper(Tvec, muBvec, muSvec, muQvec, muCvec):
            eps = 0.
            for k in range(len(self.Mass)):
                wz = self.w[k] * self.z(k, muB_div_T=muBvec,
                                        muQ_div_T=muQvec, muS_div_T=muSvec, muC_div_T=muCvec)

                def integrand(E):
                    exp_E_div_T = np.exp(-E/Tvec)
                    return wz * E**2 * (E**2-self.Mass[k]**2)**(1/2) * exp_E_div_T / (1-wz*exp_E_div_T)
                eps += self.w[k] * self.g[k] * integrateFunction(integrand, self.Mass[k], np.inf, method='quad')
            eps /= (2*np.pi**2*Tvec**4)
            return eps
        int_vec = np.vectorize(int_wrapper)
        return unvector(np.asarray(int_vec(T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T)))

    def number_density(self, T, charge='B', muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T = envector(T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T)
        if charge == 'B':
            X = self.B
        elif charge == 'Q':
            X = self.Q
        elif charge == 'S':
            X = self.S
        elif charge == 'C':
            X = self.C
        else:
            logger.TBRaise('Unrecognized charge', charge)

        def int_wrapper(Tvec, muBvec, muSvec, muQvec, muCvec):
            NX = 0.
            for k in range(len(self.Mass)):
                wz = self.w[k] * self.z(k, muB_div_T=muBvec,muQ_div_T=muQvec, muS_div_T=muSvec, muC_div_T=muCvec)
                def integrand(E):
                    exp_E_div_T = np.exp(-E/Tvec)
                    return -wz * X[k] * E * (E**2-self.Mass[k]**2)**(1/2) * exp_E_div_T / (1 - wz*exp_E_div_T)
                NX -= self.w[k] * self.g[k] * integrateFunction(integrand, self.Mass[k], np.inf, method='quad')
            NX /= (2*np.pi**2*Tvec**3)
            return NX
        int_vec = np.vectorize(int_wrapper)
        return unvector(np.asarray(int_vec(T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T)))

    def S_div_T3(self, T, muB_div_T=0., muS_div_T=0., muQ_div_T=0., muC_div_T=0.):
        """ 
        s = e + p - mu_i n_i 
        """
        T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T = envector(T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T)

        def int_wrapper(Tvec, muBvec, muSvec, muQvec, muCvec):
            S = 0.
            for k in range(len(self.Mass)):
                wz = self.w[k] * self.z(k, muB_div_T=muBvec, muQ_div_T=muQvec, muS_div_T=muSvec, muC_div_T=muCvec)

                def integrand(E):
                    exp_E_div_T = np.exp(-E/Tvec)
                    P = (E**2-self.Mass[k]**2)/3
                    eps = E**2
                    return ( wz * (E**2-self.Mass[k]**2)**(1/2) \
                                * (P + eps - Tvec*E*(self.B[k]*muBvec + self.Q[k]*muQvec + self.S[k]*muSvec + self.C[k]*muCvec)) \
                                * exp_E_div_T / (1-wz*exp_E_div_T) )

                S += self.w[k] * self.g[k] * integrateFunction(integrand, self.Mass[k], np.inf, method='quad')
            S /= (2*np.pi**2*Tvec**4)
            return S
        int_vec = np.vectorize(int_wrapper)
        return unvector(np.asarray(int_vec(T, muB_div_T, muS_div_T, muQ_div_T, muC_div_T)))


class EVHRG(HRGbase):

    """ 
    Excluded volume hadron resonance gas. b is excluded volume parameter. 
    """

    def __init__(self, Mass, g, w, B, S, Q, C=None):
        HRGbase.__init__(self, Mass, g, w, B, S, Q, C)

    def __repr__(self) -> str:
        return "EVHRG"

    def Pid(self, m, g, T):
        return g*(m/T)**2 * kn(2, (m/T)) / np.pi**2 / 2

    def baryon_pressure(self, T, b, Bi, muB_div_T=0.0, muQ_div_T=0.0, muS_div_T=0.0):

        baryon_mass = self.Mass[np.where(self.B == Bi)]
        g_baryon = self.g[np.where(self.B == Bi)]
        X_baryon = self.B[np.where(self.B == Bi)]
        X_charge = self.Q[np.where(self.B == Bi)]
        X_strange = self.S[np.where(self.B == Bi)]

        # Bi=1 Baryon pressure , Bi=-1 for anti baryon pressure
        P = []
        for k in range(len(baryon_mass)):
            P.append(self.Pid(baryon_mass[k], g_baryon[k], T))

        P = np.array(P)

        mB, mQ, mS, i, ci, bi, qi, si, temp = symbols(
            'mB, mQ , mS, i, ci, bi, qi, si, temp')

        F_pressure = (1 / (b*(T/197.3)**3)) * LambertW(
            (b*(T/197.3)**3) * Sum(Indexed('ci', i) * exp(mB * Indexed('bi', i) + mQ * Indexed('qi', i) + mS * Indexed('si', i)),
                                   (i, 0, len(P) - 1)))
        f = lambdify((mB, mQ, mS, ci, bi, qi, si, temp), F_pressure,
                     modules=['scipy', {'LambertW': lambertw}])

        pressure = f(muB_div_T/T, muQ_div_T/T, muS_div_T/T,
                     P, X_baryon, X_charge, X_strange, T).real

        return pressure

    def gen_chi(self, T, b, Bi, B_order=0, Q_order=0, S_order=0, muB_div_T=0.0, muQ_div_T=0.0, muS_div_T=0.0):

        baryon_mass = self.Mass[np.where(self.B == Bi)]
        g_baryon = self.g[np.where(self.B == Bi)]
        X_baryon = self.B[np.where(self.B == Bi)]
        X_charge = self.Q[np.where(self.B == Bi)]
        X_strange = self.S[np.where(self.B == Bi)]

        # Bi=1 Baryon pressure , Bi=-1 for anti baryon pressure
        P = []
        for k in range(len(baryon_mass)):
            P.append(self.Pid(baryon_mass[k], g_baryon[k], T))

        P = np.array(P)

        # ci = ideal gas pressure of individual particles
        mB, mQ, mS, i, ci, bi, qi, si, temp, be = symbols(
            'mB, mQ , mS, i, ci, bi, qi, si, temp, be')

        F_pressure = (1 / be) * LambertW(
            be * Sum(
                Indexed('ci', i) * exp(mB * Indexed('bi', i) + mQ *
                                       Indexed('qi', i) + mS * Indexed('si', i)),
                (i, 0, len(P) - 1)))
        expr_chi = sy.diff(F_pressure, mB, B_order, mQ, Q_order, mS, S_order)

        f = lambdify((mB, mQ, mS, ci, bi, qi, si, temp, be),
                     expr_chi, modules=['scipy', {'LambertW': lambertw}])

        chi_num = f(muB_div_T/T, muQ_div_T/T, muS_div_T/T, P,
                    X_baryon, X_charge, X_strange, T, b*(T/197.3)**3).real

        return chi_num



# 
# speedify.py                                                               
# 
# L. Altenkort, D. Clarke 
# 
# Some methods and classes to easily make python code faster. 
#

import os
import numpy as np
import concurrent.futures
from latqcdtools.base.check import checkType
import latqcdtools.base.logger as logger

HAVENUMBA=True
try:
    from numba import njit
    from numba.typed import List
except ModuleNotFoundError:
    HAVENUMBA=False

# Resolve parallelizer dependencies
DEFAULTPARALLELIZER = 'pathos.pools'
try:
    import pathos.pools
except ModuleNotFoundError:
    DEFAULTPARALLELIZER = 'concurrent.futures'


COMPILENUMBA        = False
MAXTHREADS          = os.cpu_count() 
DEFAULTTHREADS      = MAXTHREADS - 2


def numbaON():
    """ 
    Use numba wherever possible. By default it is turned off, since compilation takes some time,
    and hence you will only see a performance boost for particularly long-running functions. Must be
    called at the beginning of your code. 
    """
    global COMPILENUMBA
    COMPILENUMBA = True
    logger.info('Using numba to speed things up.')


def numbaOFF():
    """ 
    Turn off numba compilation for small functions. 
    """ 
    global COMPILENUMBA
    COMPILENUMBA = False 
    logger.info('No longer using numba.')


def compile(func):
    global COMPILENUMBA
    global HAVENUMBA
    if COMPILENUMBA and HAVENUMBA:
        logger.info('Compiling',func.__name__+'.')
        return njit(func)
    else:
        return func


def numbaList(inList):
    """ 
    Turn a list into List that numba can parse. 
    """ 
    global COMPILENUMBA
    global HAVENUMBA
    if COMPILENUMBA and HAVENUMBA:
        nList = List()
        [nList.append(x) for x in inList]
        return nList 
    else:
        return inList


class ComputationClass:

    def __init__(self, function, input_array, args, nproc, parallelizer):
        """ 
        This class contains everything needed to parallelize a function. It allows for you to
        use your favorite parallelization library and pass arguments to the function.

        Args:
            function (func): to-be-parallelized function
            input_array (array-like): run function over this array 
            nproc (int): number of processors 
            parallelizer (str): Which library should I use to parallelize?
            *add_param: Pass any additional parameters as you would to the function
        """
        checkType('int',nproc=nproc)
        checkType("array",input_array=input_array)
        checkType(str,parallelizer=parallelizer)
        self._input_array  = input_array
        self._function     = function
        self._parallelizer = parallelizer
        self._args         = args
        self._nproc        = nproc
        logger.debug('Initializing')
        logger.debug('input_array =',self._input_array)
        logger.debug('args =',self._args)
        logger.debug('parallelizer =',self._parallelizer)
        if nproc > MAXTHREADS:
            logger.warn('We recommend using fewer processes than',MAXTHREADS) 
        self._result = self.parallelization_wrapper() # compute the result when class is initialized

    def __repr__(self) -> str:
        return "ComputationClass"

    def parallelization_wrapper(self):
        if self._nproc==1:
            results = []
            for i in self._input_array:
                logger.progress(i)
                results.append(self.pass_argument_wrapper(i)) 
        else:
            if self._parallelizer=='concurrent.futures':
                with concurrent.futures.ProcessPoolExecutor(max_workers=self._nproc) as executor:
                    results=executor.map(self.pass_argument_wrapper, self._input_array)
            elif self._parallelizer=='pathos.pools':
                pool = pathos.pools.ProcessPool(processes=self._nproc)
                results = pool.map(self.pass_argument_wrapper, self._input_array)
                pool.close()
                pool.join() 
                pool.clear()
            else:
                logger.TBRaise('Unknown parallelizer',self._parallelizer)
            results = list(results)
        return results

    def pass_argument_wrapper(self, single_input):
        return self._function(single_input, *self._args)

    def getResult(self):
        return self._result


def parallel_function_eval(function, input_array, args=(), nproc=DEFAULTTHREADS, parallelizer=DEFAULTPARALLELIZER):
    """ 
    Parallelize a function over an input_array. Effectively this can replace a loop over an array and should
    lead to a performance boost.

    Args:
        function (func): to-be-parallelized function 
        input_array (array-like): array over which it should run 
        nproc (int): number of processes 

    Returns:
        array-like: func(input_array)
    """
    computer = ComputationClass(function, input_array, args=args, nproc=nproc, parallelizer=parallelizer)
    if nproc==1:
        logger.details('Using for-loop instead of',parallelizer)
    return computer.getResult()


def parallel_reduce(function, input_array, args=(), nproc=DEFAULTTHREADS, parallelizer=DEFAULTPARALLELIZER) -> float:
    """ 
    Parallelize a function over an input_array, then sum over the input_array elements. 

    Args:
        function (func): to-be-parallelized function 
        input_array (array-like): array over which it should run 
        nproc (int): number of processes 

    Returns:
        float
    """
    container = parallel_function_eval(function, input_array, nproc=nproc, args=args, parallelizer=parallelizer)
    return np.sum(container)



# 
# testing.py                                                               
# 
# D. Clarke 
# 
# Functions to assist with unit testing. 
# 

import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.base.utilities import envector
from latqcdtools.base.check import checkEqualLengths
from latqcdtools.math.math import rel_check
from latqcdtools.statistics.statistics import gaudif
from latqcdtools.base.printErrorBars import get_err_str


def print_results(res, res_true, res_err = None, res_err_true = None, text = "", prec = 1e-10, abs_prec = None):
    """ 
    Compares element-by-element the results of res with res_true. (Does the same with res_err and res_err_true,
    if you like.) Carries out with precision prec. Use abs_prec for comparisons with zero. 
    """
    test = True
    compareZero = False

    if abs_prec is not None:
        compareZero = True
    else:
        abs_prec = 1e-14

    res, res_true = envector(res, res_true)
    checkEqualLengths(res,res_true)

    if (res_err is None) and (res_err_true is not None):
        logger.TBRaise('res_err_true specified without res_err')
    if (res_err is not None) and (res_err_true is None):
        logger.TBRaise('res_err specified without res_err_true')

    if res_err is not None:
        res_err      = envector(res_err)
        res_err_true = envector(res_err_true)
        checkEqualLengths(res_err,res_err_true)

    for i in range(len(res)):
        if not rel_check(res[i], res_true[i], prec, abs_prec):
            test = False
            logger.info("res[" + str(i) + "] = " + str(res[i]) + " != res_true[" + str(i) + "] = " + str(res_true[i]))
        if res_err is not None and res_err_true is not None:
            if not rel_check(res_err[i], res_err_true[i], prec, abs_prec):
                test = False
                logger.info("res_err[" + str(i) + "] = " + str(res_err[i]) + " != res_err_true[" + str(i) + "] = " + str(res_err_true[i]))

    if test:
        return True
    else:
        if compareZero:
            logger.TBFail(text, '(abs_prec = %.2e)' % abs_prec)
        else:
            logger.TBFail(text, '(prec = %.2e)' % prec)
        return False 


def print_results_iter(res,  res_true, text):
    test = True
    res = np.asarray(res)
    res_true = np.asarray(res_true)
    it = np.nditer(res, flags=['multi_index'])
    while not it.finished:
        if not rel_check(res[it.multi_index], res_true[it.multi_index], 1e-4, 1e-6):
            test = False
            logger.info("res[" + str(it.multi_index) + "] = " + str(res[it.multi_index])
                    + " != res_true[" + str(it.multi_index) + "] = " + str(res_true[it.multi_index]))
        it.iternext()
    if test:
        return True
    else:
        logger.TBFail(text)
        return False


def gaudif_results(res, res_err, res_true, res_err_true, text = "", qcut=0.05):
    """ 
    Compares element-by-element the results of res with res_true using Gaussian difference test, i.e. it checks
    to see whether res and res_true are statistically compatible. 
    """

    test = True

    res, res_true, res_err, res_err_true = envector(res, res_true, res_err, res_err_true)

    for i in range(len(res)):

        q = gaudif(res[i], res_err[i], res_true[i], res_err_true[i])

        if q < qcut:
            test = False
            resstr     = get_err_str(res[i]     ,res_err[i])
            restruestr = get_err_str(res_true[i],res_err_true[i])
            logger.info("res["+str(i)+"] =",resstr,"!= res_true["+str(i)+"] =",restruestr,'[ q =',round(q,2),']')

    if test:
        return True 
    else:
        logger.TBFail(text)
        return False


def concludeTest(lpass):
    if lpass:
        logger.TBPass('All tests passed.')
    else:
        logger.TBError('At least one test failed.')



# 
# MILC.py
# 
# D. Clarke
# 
# Some parameter combinations and naming conventions special to projects of the MILC collaboration.
#

from latqcdtools.interfaces.interfaces import HotQCD_MILC_Params

class MILCParams(HotQCD_MILC_Params):
    """
    A class to handle and check the input parameters of a lattice run, especially for MILC.
    """

    def __repr__(self) -> str:
        return "MILCParams"



# 
# SU3.py                                                               
#
# D. Clarke
#
# Implementation of SU3 gauge group in Python. We try to implement various functions as static methods 
# so that we can compile them with numba to speed them up. We also try to leverage numpy.
# 

import numpy as np
from latqcdtools.math.math import SUN
from latqcdtools.base.speedify import compile


# Eventually we would like to use default_rng here too, but it doesn't compile straightforwardly
# using numba. Will have to think about it.
rng = np.random


@compile
def fastUnitarize(self):

    # Normalize row 0.
    norm = np.sqrt( abs(self[0,0])**2 + abs(self[0,1])**2 + abs(self[0,2])**2 )
    self[0,0] /= norm
    self[0,1] /= norm
    self[0,2] /= norm

    # Get row 1 from ON projection on row 0.
    Cre =     self[1,0].real*self[0,0].real + self[1,0].imag*self[0,0].imag \
            + self[1,1].real*self[0,1].real + self[1,1].imag*self[0,1].imag \
            + self[1,2].real*self[0,2].real + self[1,2].imag*self[0,2].imag
    Cim =     self[1,0].imag*self[0,0].real - self[1,0].real*self[0,0].imag \
            + self[1,1].imag*self[0,1].real - self[1,1].real*self[0,1].imag \
            + self[1,2].imag*self[0,2].real - self[1,2].real*self[0,2].imag
    self[1,0] -= complex( Cre*self[0,0].real - Cim*self[0,0].imag, Cre*self[0,0].imag + Cim*self[0,0].real )
    self[1,1] -= complex( Cre*self[0,1].real - Cim*self[0,1].imag, Cre*self[0,1].imag + Cim*self[0,1].real )
    self[1,2] -= complex( Cre*self[0,2].real - Cim*self[0,2].imag, Cre*self[0,2].imag + Cim*self[0,2].real )

    # Normalize row 1.
    norm = np.sqrt( abs(self[1,0])**2 + abs(self[1,1])**2 + abs(self[1,2])**2 )
    self[1,0] /= norm
    self[1,1] /= norm
    self[1,2] /= norm

    # Row 2 is vector product of rows 0 and 1.
    self[2,0] = complex( ( self[0,1].real*self[1,2].real - self[0,1].imag*self[1,2].imag ) - ( self[0,2].real*self[1,1].real - self[0,2].imag*self[1,1].imag ) ,
                        -( self[0,1].imag*self[1,2].real + self[0,1].real*self[1,2].imag ) + ( self[0,2].imag*self[1,1].real + self[0,2].real*self[1,1].imag ) )

    self[2,1] = complex( ( self[0,2].real*self[1,0].real - self[0,2].imag*self[1,0].imag ) - ( self[0,0].real*self[1,2].real - self[0,0].imag*self[1,2].imag ),
                        -( self[0,2].imag*self[1,0].real + self[0,2].real*self[1,0].imag ) + ( self[0,0].imag*self[1,2].real + self[0,0].real*self[1,2].imag ) )

    self[2,2] = complex( ( self[0,0].real*self[1,1].real - self[0,0].imag*self[1,1].imag ) - ( self[0,1].real*self[1,0].real - self[0,1].imag*self[1,0].imag ),
                        -( self[0,0].imag*self[1,1].real + self[0,0].real*self[1,1].imag ) + ( self[0,1].imag*self[1,0].real + self[0,1].real*self[1,0].imag ) )


@compile
def fastRandomize(self):
    for i in range(3):
        for j in range(3):
            self[i,j] = complex( 1 - 2*rng.uniform(0,1), 1 - 2*rng.uniform(0,1) )


class SU3(SUN):

    def __new__(cls, mat=None):
        return super(SU3, cls).__new__(cls, N=3, mat=mat)

    def __repr__(self) -> str:
        return "SU(3)"

    def su3unitarize(self):
        """ 
        Project to SU(3) using information from the first two rows. 
        """
        fastUnitarize(self)

    def setToRandom(self):
        """ 
        Turn into a randomly chosen SU(3) matrix. 
        """
        fastRandomize(self)
        self.su3unitarize()



# 
# initialize.py                                                               
# 
# D. Clarke 
# 
# Some routines to set up the toolbox, especially for keeping a record of what you did. 
# 

import os, sys, warnings
import latqcdtools.base.logger as logger
from latqcdtools.base.utilities import shell, createFilePath


INITIALIZED = False     # Global flag to check if initialization has already occurred.
DEFAULTSEED = 7271978   # Default seed for reproducibility (needed in testing). Do not Google this date.


def gitHash() -> str:
    """ 
    Obtain the current git hash. This assumes the Toolbox has been correctly installed.

    Returns:
        str: git hash 
    """
    PYTHONPATH = shell('echo $PYTHONPATH')
    try:
        for entry in PYTHONPATH.split(':'):
            # The package was called AnalysisToolbox, then LatticeToolbox, then AnalysisToolbox again. 
            if ('AnalysisToolbox' in entry) or ('LatticeToolbox' in entry):
                toolboxLocation = entry.strip()
        hash=shell('git --git-dir="'+toolboxLocation+'/.git" rev-parse HEAD').strip()
    except:
        hash='GIT_NOT_FOUND'
    return hash 


def introduceYourself():
    """ 
    Corporate branding. ASCII generated from https://patorjk.com. 
    """
    logger.info()
    logger.info(r"     _                _           _    _____           _ _                ")
    logger.info(r"    / \   _ __   __ _| |_   _ ___(_)__|_   _|__   ___ | | |__   _____  __ ")
    logger.info(r"   / _ \ | '_ \ / _` | | | | / __| / __|| |/ _ \ / _ \| | '_ \ / _ \ \/ / ")
    logger.info(r"  / ___ \| | | | (_| | | |_| \__ \ \__ \| | (_) | (_) | | |_) | (_) >  <  ")
    logger.info(r" /_/   \_\_| |_|\__,_|_|\__, |___/_|___/|_|\___/ \___/|_|_.__/ \___/_/\_\ ")
    logger.info(r"                        |___/                                             ")
    logger.info()


def initialize(logFile=None):
    """ 
    Some common tasks to do at the start of a run where you want to keep track of things. 
    """
    global INITIALIZED
    warnings.simplefilter('ignore', UserWarning)
    introduceYourself()
    if logFile is None:
        logFile = 'log/' + os.path.splitext(os.path.basename(sys.argv[0]))[0] + '.log'
    createFilePath(logFile)
    logger.createLogFile(logFile)
    logger.info("Current git commit =",gitHash())
    logger.info()
    INITIALIZED = True


def finalize():
    """ 
    Some common tasks to do when you're done. 
    """
    global INITIALIZED
    if not INITIALIZED:
        logger.warn('Called without having initialized first!')
    else:
        logger.info()
        logger.info("I'm finished!")
        logger.info()



# 
# statisticalPhysics.py                                                               
# 
# D. Clarke
# 
# A collection of methods relevant for statistical physics calculations.
#


import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.base.check import checkType
from latqcdtools.statistics.statistics import std_mean, weighted_mean, weighted_variance
from latqcdtools.base.printErrorBars import getValuesFromErrStr
from latqcdtools.base.utilities import toNumpy


def _printExponent(prefix, exponent):
    if exponent is not None:
        logger.info(prefix, round(exponent, 4))


class UniversalityClass:

    """ 
    Skeleton universality class from which all others inherit.
    """

    symm  = None
    d     = None
    alpha = None
    beta  = None
    gamma = None
    delta = None
    nu    = None
    eta   = None
    omega = None
    Tcs   = {} # J=1, no magnetic field

    def __repr__(self) -> str:
        return "UniversalityClass"

    def name(self) -> str:
        return str(self.d)+"d, "+str(self.symm)

    def exponentSummary(self):
        logger.info()
        logger.info("Summary of "+self.name()+" critical exponents:")
        _printExponent(" alpha =",self.alpha)
        _printExponent("  beta =",self.beta)
        _printExponent(" gamma =",self.gamma)
        _printExponent(" delta =",self.delta)
        _printExponent(" omega =",self.omega)
        _printExponent("    nu =",self.nu)
        logger.info()

    def hyperscalingCheck(self, tol=1e-6) -> bool:
        err1 = 2*self.beta+self.gamma-2+self.alpha
        err2 = 2*self.beta*self.delta-self.gamma-2+self.alpha
        err3 = self.nu*self.d-2+self.alpha
        if err1 > tol:
            logger.TBFail(self.name(),"fails hyperscaling check 1, err =",err1)
            return False
        if err2 > tol:
            logger.TBFail(self.name(),"fails hyperscaling check 2. err =",err2)
            return False
        if err3 > tol:
            logger.TBFail(self.name(),"fails hyperscaling check 3. err =",err3)
            return False
        return True

    def Tc(self,paper=None):
        """ Give the weighted average of literature values of Tc. These are in units with
        k_B=J=1 and no magnetic field. If there is only one known value, give that back.
        The 2d, Z(2) univerality class just gives back the Onsager solution. If you want
        to access a particular result, you can pass the DOI of the paper. """ 
        if len(self.Tcs)==0:
            logger.TBRaise('No Tc data for this universality class.')
        elif paper is not None:
            return getValuesFromErrStr(self.Tcs[paper])
        elif (self.symm=="Z_2") and (self.d==2):
            return self.Tcs['Onsager']
        elif len(self.Tcs)==1:
            for key in self.Tcs:
                return getValuesFromErrStr(self.Tcs[key])
        else:
            _Tcs, _Tces = [],[]
            for key in self.Tcs:
                m, e = getValuesFromErrStr(self.Tcs[key])
                _Tcs.append(m)
                _Tces.append(e)
            _Tcs, _Tces = toNumpy(_Tcs,_Tces)
            return weighted_mean(_Tcs, _Tces), np.sqrt(weighted_variance(_Tces))


# 3d XY model
class O2_3d(UniversalityClass):
    """ 
    3d O(2) critical exponents from JHEP08 (2016) 036 
    """
    symm  = "O(2)"
    d     = 3
    eta   = 0.03852
    nu    = 0.6719
    gamma = nu*(2-eta)
    beta  = 0.5*(nu*d-gamma)
    alpha = 2 - nu*d
    delta = (gamma+2*alpha)/(2*beta)
    def __repr__(self) -> str:
        return super().__repr__()+':'+self.name


# 3d Heisenberg model
class O3_3d(UniversalityClass):
    """ 
    3d O(3) critical exponents from JHEP08 (2016) 036 
    """
    symm  = "O(3)"
    d     = 3
    eta   = 0.0386 
    nu    = 0.7121 
    gamma = nu*(2-eta)
    beta  = 0.5*(nu*d-gamma)
    alpha = 2 - nu*d
    delta = (gamma+2*alpha)/(2*beta)
    Tcs = {
        '10.1103/PhysRevB.43.6087' : '1.44321(21)',
        '10.1103/PhysRevB.48.936'  : '1.44300(21)'
    }
    def __repr__(self) -> str:
        return super().__repr__()+':'+self.name


class O4_3d(UniversalityClass):
    """ 
    3d O(4) critical exponents from Nucl. Phys. B 675, 533-554 (2003). 
    """
    symm  = "O(4)"
    d     = 3
    beta  = 0.380
    delta = 4.824
    alpha = 2.-beta*(1.+delta)
    gamma = beta*(delta-1.)
    nu    = (beta/d)*(1+delta)
    def __repr__(self) -> str:
        return super().__repr__()+':'+self.name


# 3d Ising model
class Z2_3d(UniversalityClass):
    """ 
    3d Z_2 critical exponents from J. Stat. Phys. 157. 869-914 (2014). 
    """
    symm  = "Z_2"
    d     = 3
    nu    = 0.62999
    eta   = 0.03631
    alpha = 2 - nu*d
    gamma = nu*(2 - eta)
    beta  = (nu*d -gamma)/2
    delta = nu*d/beta-1.
    Tcs = {
        '10.1103/PhysRevB.29.4030'     : '4.51154(13)',
        '10.1103/PhysRevB.32.1720'     : '4.51162(11)',
        '10.1016/0378-4371(94)90490-1' : '4.511463(45)',
        '10.1103/PhysRevB.82.174433'   : '4.5115232(17)'
    }
    def __repr__(self) -> str:
        return super().__repr__()+':'+self.name


# 2d Ising model
class Z2_2d(UniversalityClass):
    """ 
    Exact solution for 2d Z_2 class. 
    """
    symm  = "Z_2"
    d     = 2
    alpha = 0
    beta  = 1/8
    gamma = 7/4
    delta = 15
    nu    = 1
    eta   = 1/4
    Tcs = {
        'Onsager' : 2/np.log(1+np.sqrt(2))
    }
    def __repr__(self) -> str:
        return super().__repr__()+':'+self.name


# 2d Potts q=3/ hard hexagon model 
class S3_2d(UniversalityClass):
    """ 
    Exact solution for 2d S_3 class from Baxter "Exactly Solved Models in Statistical Mechanics"
    """
    symm  = "S_3"
    d     = 2
    alpha = 1/3 
    beta  = 1/9
    gamma = 13/9 
    delta = 14
    nu    = 5/6
    eta   = 4/15
    def __repr__(self) -> str:
        return super().__repr__()+':'+self.name


# 2d Potts q=4/ Ashkin-Teller model
class S4_2d(UniversalityClass):
    """ 
    Exact solution for 2d S_4 class from Baxter, "Exactly Solved Models in Statistical Mechanics"
    """
    symm  = "S_4"
    d     = 2
    alpha = 2/3 
    beta  = 1/12
    gamma = 7/6
    delta = 15
    nu    = 2/3
    eta   = 1/4
    def __repr__(self) -> str:
        return super().__repr__()+':'+self.name


def reweight(X, pRW, p0, S):
    """ 
    Reweight an observable X computed at a simulation point p0 to a nearby
    simulation point pRW. We assume the action depends linearly on the simulation
    parameter, i.e. S' ~ p S

    By the way, if you are going to pair this with a jackknife to estimate e.g.
    where a response function is maximized, make sure you pick enough reweighting
    points to properly resolve where the maximum is.

    Args:
        X (np.array): Measurements to reweight. 
        pRW (float): Reweight to this target. 
        p0 (float): Simulation point.
        S (np.array): Measurements of the action (extensive) divided by parameter p. 
    """
    checkType(np.ndarray,X=X)
    checkType(np.ndarray,S=S)
    arg = (pRW-p0)*S
    # Subtracting by np.max(arg) is a trick to avoid overflow and improve
    # numerical stability.
    Z_i = np.exp( arg-np.max(arg) ) 
    Z   = np.sum(Z_i)
    return np.sum(X*Z_i/Z)




#
# statistics.py
#
# D. Clarke, H Sandmeyer 
#
# A collection of basic methods for statistical analysis. The following methods have been adapted from software of
# Bernd Berg, Markov Chain Monte Carlo and their Statistical Analysis, World Scientific, 2004, ISBN=978-981-3106-37-6:
#     gaudif, studif
#


import numpy as np
import scipy as sp
from latqcdtools.math.num_deriv import diff_jac 
from latqcdtools.math.math import logDet, normalize, invert, isVector, checkVector
from latqcdtools.base.plotting import fill_param_dict, plot_fill, plot_lines, plt, _initializePlt,\
    _add_optional, _getAxObject, ZOD, _update_handles, _update_labels, set_params
from latqcdtools.base.utilities import toNumpy, appendToDocstring, unvector, createFilePath
from latqcdtools.base.cleanData import clipRange
from latqcdtools.base.check import checkType, checkEqualLengths
from latqcdtools.base.readWrite import writeTable
import latqcdtools.base.logger as logger
import matplotlib


NUMPYCOMMENT = """ 
    The default behavior of numpy is to flatten the data, flagged by axis=None. This is
    something that is never needed in our context. Changing the default to axis=0 means
    applying this function to an np.ndarray of shape (N,M) yields an array of shape (M,). 
    """

ICARGCOMMENTS = """
        xdata (array-like)
        ydata (array-like)
        cov (array-like): covariance matrix 
        func (func)
        args (tuple, optional): arguments to func. Defaults to ().
        params (tuple, optional): model parameters. Defaults to ().
        prior (array-like, optional): Bayesian priors. Defaults to None.
        priorsigma (array-like, optional): Bayesian prior errors. Defaults to None.
    """


def meanArgWrapper(func,used_data,args):
    if isinstance(args, dict):
        return func(used_data, **args)
    else:
        return func(used_data, *args)


@appendToDocstring(NUMPYCOMMENT)
def std_median(data, axis = 0):
    """ 
    Compute the median. 
    """
    return np.median(data, axis)


@appendToDocstring(NUMPYCOMMENT)
def std_mean(data, axis = 0):
    """ 
    Compute the mean. 
    """
    return np.mean(data, axis)


@appendToDocstring(NUMPYCOMMENT)
def std_var(data, axis = 0):
    """ 
    Calculate unbiased (ddof = 1) estimator for the variance. 
    """
    data = np.asarray(data)
    return np.var(data, axis = axis, ddof = 1)


@appendToDocstring(NUMPYCOMMENT)
def std_dev(data, axis = 0):
    """ 
    Calculate unbiased (ddof = 1) estimator for the standard deviation. 
    """
    data = np.asarray(data)
    return np.std(data, axis = axis, ddof = 1)


@appendToDocstring(NUMPYCOMMENT)
def std_err(data, axis = 0):
    """ 
    Standard deviation of the sample mean according to the CLT. 
    """
    data = np.asarray(data)
    return std_dev(data, axis) / np.sqrt(data.shape[axis])


def expandArgs(func, x, params=(), args=()):
    """ 
    In general we distinguish between parameters and arguments. Parameters should be passed
    together as a collection, e.g. as a tuple, list, or np.array. Other function arguments can
    be passed how you like and will be expanded here.

    Args:
        func (func)
        x (array-like)
        params (tuple, optional): Model parameters. Defaults to ().
        args (tuple, optional): Function arguments. Defaults to ().

    Returns:
        func(x,params,args) 
    """
    if len(params)==0:
        # Python treats func(x,*()) as func(x)
        return func(x, *args)
    else:
        return func(x, params, *args)


def checkDomain(domain):
    """ 
    Some methods require that you do something over an interval, which we refer to in this module as
    a 'domain'. This checks the domain makes sense.

    Args:
        domain (tuple)
    """
    checkType(tuple,domain=domain)
    if len(domain) != 2:
        logger.TBRaise('A domain is a tuple of the form (xmin,xmax) specifying the interval [xmin,xmax].',frame=3)
    if domain[0]>=domain[1]:
        logger.TBRaise('Must have domain[1]>domain[0].',frame=3)


def checkPrior(prior,priorsigma):
    """ 
    Make sure prior and priorsigma status are compatible. 
    """
    if prior is None and priorsigma is not None:
        logger.TBRaise('prior = None, priorsigma != None',frame=3)
    if priorsigma is None and prior is not None:
        logger.TBRaise('prior != None, priorsigma = None',frame=3)


def checkTS(ts):
    """ 
    Some methods require 1-d time series. This checks that the type, dimensionality,
    and length are appropriate.

    Args:
        ts (array-like): time series 
    """
    checkVector(ts)
    if len(ts) < 2:
        logger.TBRaise('Time series needs at least two measurements.',frame=3)


def checkProb(p):
    checkType('real',p=p)
    if not 0 <= p <= 1:
        logger.TBRaise('Probabilities must be between 0 and 1.',frame=3)


def countParams(func,params) -> int:
    """ 
    Count number of model parameters. For a typical function without priors,
    we count the length of the params array. Otherwise we assume it's a spline.

    Args:
        func (func)
        params (array-like): model parameters.
        priorsigma (array-like, optional): Bayesian prior errors. Defaults to None.

    Returns:
        int: number of parameters. 
    """
    nparam = len(params)
    if nparam == 0:
        try:
            # customSpline
            nparam = len(func.get_coeffs())
            # CubicSpline
        except AttributeError:
            nparam = len(func.c)
    return nparam


def countPriors(priorsigma=None) -> int:
    """ 
    The number of priors is the number of finite prior error bars.

    Args:
        priorsigma (array-like, optional): Bayesian prior errors. Defaults to None.

    Returns:
        int: Number of priors 
    """
    nprior = 0
    if priorsigma is not None:
        nprior = np.count_nonzero(priorsigma!=np.inf)
    return nprior


def DOF(ndat,nparam,priorsigma=None) -> int:
    """  
    Compute the number of degrees of freedom. Depends on whether you use priors. Any input priors are taken as
    initial guesses for the fit algorithm. If you would like parameters in the prior array to be treated as a 
    starting guess only, and not as a Bayesian prior, set its corresponding error to np.inf. Hence when there
    are priors, the number of degrees of freedom equals the number of ydata, less the number of finite prior errors.

    Args:
        ndat (int): number of data 
        nparam (int): number of model parameters 
        priorsigma (array-like, optional): Bayesian prior errors. Defaults to None.

    Returns:
        int: number of degrees of freedom 
    """
    checkType('int',ndat=ndat)
    checkType('int',nparam=nparam)
    nprior = countPriors(priorsigma) 
    dof = ndat + nprior - nparam 
    logger.debug('dof =',dof,'ndat =',ndat,'nparam =',nparam,'nprior =',nprior)
    if dof < 0:
        logger.TBRaise('ndat =',ndat,'nparam =',nparam,'nprior =',nprior)
    return dof


@appendToDocstring(args=ICARGCOMMENTS,returns='\n        float: chi^2\n')
def chisquare(xdata, ydata, cov, func, args=(), params=(), prior=None, priorsigma=None) -> float:
    """ 
    Calculate chi^2, see e.g. eq. (8.28) of Sivia and Skilling or eq. (A1) of
    10.1103/PhysRevD.90.054506. We assume priors are not correlated with data.
    """
    checkPrior(prior,priorsigma)
    y    = expandArgs(func,xdata,params,args)
    diff = ydata-y
    res  = diff @ invert(cov) @ diff
    if prior is not None:
        res += np.sum((np.array(params) - prior)**2 / priorsigma**2)
    return res


@appendToDocstring(args=ICARGCOMMENTS,returns='\n        float: log( Gaussian Bayes factor )\n' ) 
def logGBF(xdata, ydata, cov, func, args=(), params=(), prior=None, priorsigma=None) -> float:
    """ 
    log P(data|model). This quantity is useful for comparing fits of the same data to different models that
    have different priors and/or fit functions. The model with the largest logGBF is the one preferred by the data.
    Differences in logGBF smaller than 1 are not very significant. Gaussian statistics are assumed.
    """
    chi2   = chisquare(xdata, ydata, cov, func, args, params, prior, priorsigma)
    nparam = countParams(func,params)
    dof    = DOF(len(ydata),nparam,priorsigma)
    if prior is None:
        return 0.5*( - logDet(cov) - chi2 - dof*np.log(2*np.pi) )
    else:
        return 0.5*( - logDet(cov) - chi2 - dof*np.log(2*np.pi) + logDet(np.diag(clipRange(priorsigma)**2)) )


@appendToDocstring(args=ICARGCOMMENTS,returns='\n        float: AIC\n')
def AIC(xdata, ydata, cov, func, args=(), params=(), prior=None, priorsigma=None) -> float:
    """ 
    The Akaike information criterion (AIC) is a measure of how well a fit performs. It builds on the likelihood
    function by including a penalty for each d.o.f. This is useful in a context where you have multiple models to
    choose from,and hence different numbers of d.o.f. possible. It's also useful when you are worried about
    overfitting. The preferred model minimizes the AIC.
    """
    nparam     = countParams(func,params)
    likelihood = logGBF(xdata, ydata, cov, func, args, params, prior, priorsigma)
    return 2*nparam - 2*likelihood


@appendToDocstring(args=ICARGCOMMENTS,returns='\n        float: corrected AIC\n')
def AICc(xdata, ydata, cov, func, args=(), params=(), prior=None, priorsigma=None) -> float:
    """ 
    Corrected AIC (AICc). When the sample size is smaller, it increases the chance AIC will select a model with too
    many parameters. The AICc tries to further correct for this. In the limit that the number of data points goes to
    infinity, one recovers the AIC.
    """
    nparam = countParams(func,params)
    nprior = countPriors(priorsigma) 
    ndat   = len(ydata) + nprior
    aic    = AIC(xdata, ydata, cov, func, args, params, prior, priorsigma)
    return aic + 2*(nparam**2+nparam)/(ndat-nparam+1)


def BAIC(xdata, ydata, cov, func, args=(), params=(), Ncut=0, modelPrior=1) -> float:
    """ 
    Bayesian Akaike information criterion of 2208.14983. It uses the chi^2 as its likelihood
    function and includes penalties for having many fit parameters and cutting many data from
    your original sample.

    Args:
        xdata (array-like)
        ydata (array-like)
        cov (array-like): covariance matrix 
        func (func)
        args (tuple, optional): arguments to func. Defaults to ().
        params (tuple, optional): model parameters. Defaults to ().
        prior (array-like, optional): Bayesian priors. Defaults to None.
        priorsigma (array-like, optional): Bayesian prior errors. Defaults to None.
        Ncut (int, optional): The number of data trimmed from your fit. Defaults to 0.
        modelPrior (float, optional): The prior probability of your model. Defaults to 1.
        
    Returns:
        float: Bayesian AIC 
    """
    nparam   = countParams(func,params)
    chi2data = chisquare(xdata, ydata, cov, func, args, params, None, None)
    logger.debug('chi_data^2 =',chi2data)
    return -2*np.log(modelPrior) + chi2data + 2*nparam + 2*Ncut 


def pearson(x,y) -> float:
    """ 
    Get the Pearson correlation coefficient between the time series x and y.

    Args:
        x (array-like)
        y (array-like)

    Returns:
        float: R 
    """
    checkTS(x)
    checkTS(y)
    checkEqualLengths(x,y)
    return np.corrcoef(x,y)[0,1]


def covariance(x,y) -> float:
    """ 
    Unbiased estimator of the covariance between the time series x and y.

    Args:
        x (array-like)
        y (array-like)

    Returns:
        float: cov 
    """
    checkTS(x)
    checkTS(y)
    checkEqualLengths(x,y)
    return np.cov(x,y,ddof=1)[0,1]


def weighted_mean(data, err) -> float:
    """ 
    Compute the weighted mean. Here the weights are Gaussian error bars.
    See e.g. https://ned.ipac.caltech.edu/level5/Leo/Stats4_5.html.

    Args:
        data (array-like)
        errcov (array-like): error if 1d or cov if 2d. 

    Returns:
        float: weighted mean 
    """
    checkType(np.ndarray,data=data) 
    checkType(np.ndarray,err=err) 
    weights  = 1/err**2
    return np.sum( weights @ data )/np.sum(weights)


def weighted_variance(err) -> float:
    """ 
    Get variance of above weighted mean, when the weights are statistical errors. 

    Args:
        err (array-like)

    Returns:
        float: weighted variance 
    """
    checkType(np.ndarray,err=err) 
    weights = 1/err**2
    return 1/np.sum(weights)


def biased_sample_variance(data, err) -> float:
    """ 
    Compute the biased weighted sample variance, i.e. the biased variance of an 
    individual measurement and not the variance of the mean.

    Args:
        data (array-like)
        err (array-like)

    Returns:
        float: sample variance 
    """
    checkType(np.ndarray,data=data) 
    checkType(np.ndarray,err=err) 
    mean = weighted_mean(data, err)
    weights = 1/err**2
    V1 = np.sum(weights)
    return weights.dot((data - mean)**2) / V1


def unbiased_sample_variance(data, err) -> float:
    """ 
    Compute the unbiased weighted sample variance, i.e. the unbiased variance of an individual measurement and not
    the variance of the mean. Do not use this function if your weights are frequency weights. 
    """
    checkType(np.ndarray,data=data) 
    checkType(np.ndarray,err=err) 
    weights = 1/err**2
    V1 = np.sum(weights)
    V2 = np.sum(weights**2)
    return biased_sample_variance(data, weights) / ( 1 - (V2 / V1**2))


def unbiased_mean_variance(data, err) -> float:
    """ 
    Compute the unbiased variance of a weighted mean. Do not use this function if your weights are frequency
    weights. This is more like a systematic error. The absolute size of the weights does not matter. The error is
    constructed using the deviations of the individual data points. 
    """
    checkType(np.ndarray,data=data) 
    checkType(np.ndarray,err=err) 
    weights = 1/err**2
    V1 = np.sum(weights)
    V2 = np.sum(weights**2)
    return biased_sample_variance(data, weights) * V2 / ( V1**2 - V2)


def cov_to_cor(cov) -> np.ndarray:
    """ 
    Normalize covariance matrix to create correlation matrix.

    Args:
        cov (np.ndarray)

    Returns:
        np.ndarray: correlation matrix 
    """
    checkType(np.ndarray,cov=cov) 
    diagonal_sqrt = np.sqrt(np.diag(cov))
    return cov / np.outer(diagonal_sqrt, diagonal_sqrt)


def confidence_ellipse(x,y,CI=None,**params):
    """ 
    Plot a confidence ellipse according to the data x, y. The confidence is only meaningful 
    assuming the x and y are Gaussian distributed. By default, draws an ellipse that captures
    roughly 39% of the data.

    Args:
        x (array-like)
        y (array-like)
        ax (matplotlib ax object)
        color (str, optional): Color of the ellipse edge. Defaults to 'r'.
        C (float, optional): Desired confidence. Defaults to ~0.39.

    Returns:
        float, float: semi-major and semi-minor lengths of drawn ellipse 
    """
    checkType(np.ndarray,x=x)
    checkType(np.ndarray,y=y)
    checkEqualLengths(x,y)

    _initializePlt(params)
    optional = _add_optional(params)
    ax = _getAxObject(params)

    ZOD = params['ZOD']
    if ZOD is None:
        ZOD = globals()['ZOD']

    if params['color'] is not None:
        color=params['color']
    else:
        color='r'

    if params['markerfill']:
        markerfill=params['color']
    else:
        markerfill="None"

    if CI is None:
        s = 1
    else:
        if CI<0 or CI>1:
            logger.TBRaise('Confidence limits are between 0 and 1.') 
        s = np.sqrt(-2*np.log(1-CI))

    data = np.vstack((x, y))
    cov  = np.cov(data)
    eigvals, eigvecs = np.linalg.eig(cov)
    maj_eigvec = eigvecs[:,np.argmax(eigvals)]
    theta = np.rad2deg( np.arctan2(maj_eigvec[1],maj_eigvec[0]) )
    a = s*np.sqrt(np.max(eigvals))
    b = s*np.sqrt(np.min(eigvals))
    ellipse = matplotlib.patches.Ellipse((std_mean(x), std_mean(y)), width = 2*a, height = 2*b, angle = theta,
                                          edgecolor=color, fc=markerfill, zorder=ZOD, alpha=params['alpha'],
                                          **optional)
    ax.add_patch(ellipse)

    if params['label'] is not None:
        _update_labels(ax,params['label'])
        _update_handles(ax,ellipse)
        set_params(**params) # Needed to put in labels

    globals()['ZOD'] += 1

    return a, b, theta 


def plot_correlation(mat,ax=plt):
    """ Plot correlation matrix as a heatmap.

    Args:
        mat (np.ndarray): correlation matrix
        ax (matplotlib ax object): Defaults to plt.
    """
    checkType(np.ndarray,mat=mat)
    heatmapColors = ['blue', 'white', 'red']
    cmap = matplotlib.colors.LinearSegmentedColormap.from_list('custom_cmap', heatmapColors)
    ax.imshow(mat, cmap=cmap, vmin=-1,vmax=1)
    ax.colorbar()


def forcePositiveSemidefinite(mat):
    """ Doctors a noisy correlation matrix mat to be positive semidefinite if it isn't already. 
    Uses algorithm of Rebonato and Jaeckel, DOI: 10.2139/ssrn.1969689
    
    Args:
        mat (np.ndarray)

    Returns:
        np.ndarray: positive semidefinite matrix 
    """
    eigvals, S = np.linalg.eig(mat)
    eigvals[eigvals<0] = 0
    sqrtLambda = np.diag(np.sqrt(eigvals))

    T = np.sum(S**2*eigvals,axis=1)
    sqrtT = np.diag(np.sqrt(1/T))
    B = sqrtT @ S @ sqrtLambda

    return B @ B.T


def dev_by_dist(data, axis=0, return_both_q=False, percentile=68):
    """ 
    Calculate the distance between the median and 68% quantiles. Returns the larger of the two 
    distances. This method is used sometimes to estimate error, for example in the bootstrap. 
    """
    data = np.asarray(data)
    median = np.nanmedian(data, axis)
    numb_data = data.shape[axis]
    idx_dn = max(int(np.floor((numb_data-1) / 2 - percentile/100 * (numb_data-1) / 2)), 0)
    idx_up = min(int(np.ceil((numb_data-1) / 2 + percentile/100 * (numb_data-1) / 2)), numb_data-1)
    #TODO: Need to handle the situation idx_dn=idx_up a bit carefully. the SRI project presents
    # some data with this issue, and you can troubleshoot that.
    sorted_data = np.sort(data - np.expand_dims(median, axis), axis=axis)
    q_l = np.take(sorted_data, idx_dn, axis)
    q_r = np.take(sorted_data, idx_up, axis)
    if return_both_q:
        return np.abs(q_l), np.abs(q_r)
    else:
        return np.max(np.stack((np.abs(q_l), np.abs(q_r)), axis=0), axis=0)


def error_prop(func, means, errors, grad=None, args=()):
    """
    Use error propagation to propagate some errors through function func. The function should have the form
        func( data ), 
    where data is a 1-d array of input variables. 

    Args:
        func (func)
        means (np.ndarray)
        errors (np.ndarray)
        grad (func, optional): Gradient function. Defaults to None.
        args (tuple, optional): Arguments of func. Defaults to ().

    Returns:
        np.ndarray, np.ndarray: f, f_err 
    """
    checkType(np.ndarray,errors=errors)
    checkVector(means)
    mean = func(means, *args)

    # Test if we got a covariance matrix
    if isVector(errors): 
        errors = np.diag(errors**2)

    if type(mean) is tuple:
        logger.TBRaise("Tuples are not supported for error propagation")

    if grad is not None:
        grad = grad(means, *args)
    else:
        grad = diff_jac(means, func, args).T

    error = 0
    try:
        for i in range(len(grad)):
            for j in range(len(grad)):
                error += grad[i] * grad[j] * errors[i, j]
        error = np.sqrt(error)
    except TypeError:
        error += abs(grad * errors[0])
    return unvector(mean), unvector(error)


def error_prop_func(x, func, params, params_err, grad=None, args=()):
    """ 
    Propagate error in f(x;params,params_err). This needs its own special treatment, since
    the error propagation method on its own only propagates params_err to f(params,params_err).

    Args:
        x (array-like)
        func (func)
        params (array-like): Model parameters. 
        params_err (array-like): Error in model parameters. 
        grad (func, optional): Gradient function. Defaults to None.
        args (tuple, optional): Arguments of func. Defaults to ().
    """
    def wrap_func(m,a=()):
        if isinstance(a,tuple): 
            return func(x,m,*a)
        else:
            return func(x,m,a)
    if grad is None:
        wrap_grad = None
    else:
        def wrap_grad(m,a=()):
            if isinstance(a,tuple): 
                return grad(x,m,*a)
            else:
                return grad(x,m,a)
    return error_prop(wrap_func, params, params_err, wrap_grad, args)[1]


def gaudif(x1,e1,x2,e2) -> float:
    """ 
    Likelihood that difference between outcomes x1 and x2 is due to chance, assuming x1 and x2 are
    both drawn from a normal distribution with the same mean. A rule of thumb is that this is more
    appropriate when one estimated x1 and x2 using ~30 or more measurements.

    Args:
        x1 (float): mean 1 
        e1 (float): standard error 1
        x2 (float): mean 2
        e2 (float): standard error 2

    Returns:
        float: p-value 
    """
    if e1<0 or e2<0:
        logger.TBRaise('Error bars should be non-negative. Got',e1,e2)
    sigma = np.sqrt(e1**2 + e2**2)
    z     = abs(x1-x2)/(sigma * np.sqrt(2.))
    return 1.0 - sp.special.erf(z)


def studif(x1,e1,ndat1,x2,e2,ndat2) -> float:
    """ 
    Likelihood that difference between outcomes x1 and x2 is due to chance, assuming x1 and x2 are
    both drawn from a normal distribution with the same mean. A rule of thumb is that this is more
    appropriate when one estimated x1 and x2 using ~30 or fewer measurements. Of course, you can
    always compare this with gaudif to get a better idea.

    Args:
        x1  (float): mean 1 
        e1  (float): standard error 1
        ndat1 (int): number of measurements used to compute x1
        x2  (float): mean 2
        e2  (float): standard error 2
        ndat2 (int): number of measurements used to compute x2

    Returns:
        float: p-value 
    """
    if e1<0 or e2<0:
        logger.TBRaise('Error bars should be non-negative. Got',e1,e2)
    if ndat1<2 or ndat2 <2:
        logger.TBRaise('Need at least 2 data. Got',ndat1,ndat2)
    dof   = ndat1 + ndat2 - 2
    var12 = ndat1*e1**2
    var22 = ndat2*e2**2
    sigma = np.sqrt( (1/ndat1+1/ndat2)*( (ndat1-1)*var12 + (ndat2-1)*var22 )/dof )
    t     = abs(x1-x2)/sigma
    x     = dof/(dof+t**2)
    if t==0:
        return 1
    elif t>0:
        return sp.special.betainc(dof/2,1/2,x)
    else:
        return 2-sp.special.betainc(dof/2,1/2,x)


def goodnessOfFit(dof, chi2) -> float:
    """ 
    The q-value or goodness of fit.

    Args:
        dof (int): number of degrees of freedom 
        chi2 (float): the chi2 

    Returns:
        float: Q 
    """
    checkType("int",dof=dof)
    return sp.special.gammaincc(dof/2,chi2/2)


def plot_func(func, domain, params=(), args=(), func_err=None, params_err=(), 
              grad = None, swapXY=False, npoints=1000, **kwargs):
    """ 
    Plot a function along with its error bands.

    Args:
        func (func)
        domain (tuple): Domain of function. 
        params (tuple, optional): Model parameters. Defaults to ().
        params_err (tuple, optional): Error in model parameters. Defaults to ().
        args (tuple, optional): Optional function arguments. Defaults to ().
        func_err (func, optional): Explicit error function. Defaults to None.
        grad (func, optional): Explicit function gradient to compute error. Defaults to None.
        swapXY (bool, optional): Swap X and Y variables in plot. Defaults to False.
        npoints (int, optional): Number of points to use for plotting. Defaults to 1000.
    """
    checkDomain(domain)
    checkType('int',npoints=npoints)
    fill_param_dict(kwargs)
    kwargs['marker'] = None
    xmin = domain[0] 
    xmax = domain[1] 

    xdata = np.arange(xmin, xmax, (xmax - xmin) / npoints)
    try:
        ydata = func(xdata, params, *args)
    except TypeError:
        ydata = func(xdata, *params, *args)

    # Received an explicit error function:
    if func_err is not None:
        ydata_err = func_err(xdata, params, params_err, *args)
        if swapXY:
            return plot_fill(xdata, ydata, yedata=None, xedata=ydata_err, center=False, **kwargs)
        else:
            return plot_fill(xdata, ydata, ydata_err, center=False, **kwargs)

    # No explicit error function, but received a covariance matrix: 
    elif len(params_err) > 0:
        ydata_err = error_prop_func(xdata, func, params=params, params_err=params_err, grad=grad, args=args)
        if swapXY:
            return plot_fill(xdata, ydata, yedata=None, xedata=ydata_err, center=False, **kwargs)
        else:
            return plot_fill(xdata, ydata, ydata_err, center=False, **kwargs)

    # No errors at all: 
    else:
        if swapXY:
            return plot_lines(ydata, xdata, yedata=None, xedata=None, **kwargs)
        else:
            return plot_lines(xdata, ydata, yedata=None, xedata=None, **kwargs)


def save_func(func, domain, params=(), args=(), func_err=None, params_err=(), 
              grad = None, npoints=1000, header=None, filename='func.d', **kwargs):
    """ 
    Save a function along with its error bands.

    Args:
        func (func)
        domain (tuple): Domain of function. 
        params (tuple, optional): Model parameters. Defaults to ().
        params_err (tuple, optional): Error in model parameters. Defaults to ().
        args (tuple, optional): Optional function arguments. Defaults to ().
        func_err (func, optional): Explicit error function. Defaults to None.
        grad (func, optional): Explicit function gradient to compute error. Defaults to None.
        swapXY (bool, optional): Swap X and Y variables in plot. Defaults to False.
        npoints (int, optional): Number of points to use for plotting. Defaults to 1000.
    """
    checkDomain(domain)
    checkType('int',npoints=npoints)
    checkType(str,filename=filename)
    xmin = domain[0] 
    xmax = domain[1] 

    createFilePath(filename)

    xdata = np.arange(xmin, xmax, (xmax - xmin) / npoints)
    try:
        ydata = func(xdata, params, *args)
    except TypeError:
        ydata = func(xdata, *params, *args)

    if func_err is not None:
        ydata_err = func_err(xdata, params, params_err, *args)
        writeTable(filename,xdata,ydata,ydata_err,header=header,**kwargs)

    elif len(params_err) > 0:
        ydata_err = error_prop_func(xdata, func, params=params, params_err=params_err, grad=grad, args=args)
        writeTable(filename,xdata,ydata,ydata_err,header=header,**kwargs)

    else:
        writeTable(filename,xdata,ydata,header=header,**kwargs)


def getModelWeights(IC) -> np.ndarray:
    """ 
    Convert information criteria IC to normalized probability weights.

    Args:
        IC (array-like): Array of information criteria 

    Returns:
        np.array: Probability weights 
    """
    checkType(np.ndarray,IC=IC)
    return normalize(np.exp(-0.5*IC),p=1)


def modelAverage(data,err,IC,return_syst=False):
    """ 
    Given some fit results, corresponding error, and information criteria, compute
    a weighted model average.

    Args:
        data (array-like): Fit results 
        err (array-like): Result errors 
        IC (array-like): Information criteria 

    Returns:
        tuple: Model average and error (optionally systematic error)
    """
    checkType(np.ndarray,data=data)
    checkType(np.ndarray,err=err)
    checkType(np.ndarray,IC=IC)
    checkEqualLengths(data,err,IC)
    pr = getModelWeights(IC)
    mean = np.sum(pr*data)
    var_stat = np.sum(pr*err**2)
    var_syst = np.sum(pr*data**2) - mean**2
    var = var_stat + var_syst 
    if return_syst:
        return mean, np.sqrt(var), np.sqrt(var_syst)
    else:
        return mean, np.sqrt(var)


def empiricalCDF(data):
    """ 
    Create the x and y coordinates needed to plot the empirical CDF
    of a 1-d set of data.

    Args:
        data (array-like): measurements 

    Returns:
        func: CDF 
    """
    checkTS(data)
    return sp.stats.ecdf(data).cdf.evaluate


def KSTest_2side(data1,data2) -> float:
    """ 
    2-sided Kolmogorov test. Gives back the likelihood that the observed difference between
    data1 and data2 are at least as extreme as suggested by the Kolmogorov statistic.

    Args:
        data1 (array-like)
        data2 (array-like)

    Returns:
        float: 1-p 
    """
    checkType(np.ndarray,data1=data1)
    checkType(np.ndarray,data2=data2)
    data1,data2 = toNumpy(data1,data2)
    return 1 - sp.stats.kstest(data1, data2).pvalue 


def KSTest_1side(data,cdf) -> float:
    """ 
    1-sided Kolmogorov test. Gives back the likelihood that the observed difference between
    data and cdf are at least as extreme as suggested by the Kolmogorov statistic.

    Args:
        data (array-like)
        cdf (function)

    Returns:
        float: 1-p 
    """
    checkType(np.ndarray,data=data)
    return 1 - sp.stats.kstest(data, cdf).pvalue


def binSeries(data,nbins) -> np.ndarray:
    """ 
    Take a time series and bin it. Bin 0 is the average over the first binsize elements,
    bin 1 the average over the next binsize elements, and so on.

    Args:
        data (array-like)
        nbins (int)

    Returns:
        np.ndarray: Binned data
    """
    checkTS(data)
    checkType('int',nbins=nbins)
    ndat=len(data)
    binsize=int(ndat/nbins)
    reshaped_data = data[:binsize * nbins].reshape(nbins, binsize)
    return np.apply_along_axis(std_mean, 1, reshaped_data)


def symmetrizeError(lo,hi,central,method='conservative') -> float:
    checkType(str,method=method)
    checkType("real",lo=lo)
    checkType("real",hi=hi)
    checkType("real",central=central)
    if (lo<=0) or (hi<=0):
        logger.TBRaise('Uncertainties must be positive. Got lo, hi =',lo,hi)
    if method=='conservative':
        err  = np.max([lo,hi])
        mean = central
    # FLAG method from 10.1140/epjc/s10052-022-10536-1
    elif method=='FLAG':
        if lo>hi:
            mean = central + (lo-hi)/4
            err  = (hi + 3*lo)/4
        else:
            mean = central + (hi-lo)/4
            err  = (lo + 3*hi)/4
    else:
        logger.TBRaise('Unrecognized method',method)
    return mean, err



# staticPotential.py
# 
# D. Clarke, K. Ebira
# 
# Some methods related to the static potential in lattice QCD.
# Optimized GPU implementation for improved scaling with lattice size.

import numpy as np
from numba import cuda
import latqcdtools.base.logger as logger
from latqcdtools.base.speedify import numbaON, compile

numbaON()

#------------------------------------------------------------------------------
# Static potential functions
#------------------------------------------------------------------------------
def V_Teq0(r) -> float:
    """ 
    Zero temperature quark potential in [MeV], takes r in [fm]. The parameters a, b, and c come from
    a Levenberg-Marquardt fit of the data in Fig 14 of Phys. Rev. D90 (2014) 094503. These numbers can
    be obtained again by running analysistoolbox/hisq_potential/fit_hisq_pot.py. 
    """
    #    result:  [ -91.30436191 1022.25286821  106.70659264]
    #     error:  [0.53809612 2.51598869 2.58370288]
    #  chi2/dof:  0.8083127937775374
    a =  -91.30436191
    b = 1022.25286821
    c =  106.70659264
    return a/r + b*r + c



def fitV_Teq0(r, a, b, c) -> float:
    """ 
    Fit form of standard Cornell potential. Fit to be done in lattice units.
    """
    return a + b/r + c*r


def fitV_Teq0_oneloop(r,a,b,c,d) -> float:
    """ 
    Including one-loop corrections to Coulomb. See Nucl. Phys. B 129 (1977) and Phys. Lett B 92 (1980).
    Fit to be done in lattice units.
    """
    return a + ( b + d*np.log(r) )/r + c*r


def fitV_Teq0_twoloop(r,a,b,c,d,e) -> float:
    """ 
    Including two-loop corrections to Coulomb. See Nucl. Phys. B 501 (1997). Fit to be done in lattice units.
    """
    return a + ( b + d*np.log(r) + e*np.log(np.log(r)) )/r + c*r


#------------------------------------------------------------------------------
# CUDA utility functions
#------------------------------------------------------------------------------
def get_optimal_block_size():
    """
    Returns an optimal block size based on the current CUDA device.
    Defaults to 256 if device information cannot be obtained.
    
    Returns:
        int: Optimal threads per block
    """
    try:
        device = cuda.get_current_device()
        # Max threads per block is device.MAX_THREADS_PER_BLOCK, but usually better to use less
        return min(256, device.MAX_THREADS_PER_BLOCK)
    except:
        return 256


#------------------------------------------------------------------------------
# CUDA kernels
#------------------------------------------------------------------------------
@cuda.jit
def compute_sine_terms_kernel(sinf, cw, Ns, sine_terms_lookup):
    """
    Precompute sin terms for all k1, k2, k3 combinations to improve performance.
    
    Args:
        sinf: Precomputed sine values
        cw: Coefficient for improved action
        Ns: Spatial extension of lattice
        sine_terms_lookup: Output array for precomputed sine terms
    """
    # Get thread position
    idx = cuda.grid(1)
    
    # Calculate 3D indices
    total_indices = Ns**3
    if idx >= total_indices:
        return
    
    k3 = idx % Ns
    temp = idx // Ns
    k2 = temp % Ns
    k1 = temp // Ns
    
    if not (k1 + k2 + k3 == 0):
        # Calculate sine term once and store it
        r2 = 0.0
        for k in [k1, k2, k3]:
            sin_k = sinf[k]
            r2 += sin_k*sin_k + cw*sin_k*sin_k*sin_k*sin_k
        
        # Store result in 3D lookup table
        sine_terms_lookup[k1, k2, k3] = 1.0 / (4.0 * r2)
    else:
        sine_terms_lookup[k1, k2, k3] = 0.0

@cuda.jit
def calculate_potential_kernel(points, Ns, cosf, sine_terms_lookup, point_results):
    """
    CUDA kernel for calculating lattice QCD potentials for a set of spatial points.
    
    This kernel calculates the static potential for each input point. Each block 
    processes one spatial point, with the workload distributed across threads within the block.
    
    Args:
        points: Array of spatial coordinates (x,y,z) with shape (N,3)
        Ns: Size of the lattice in each dimension
        cosf: Precomputed cosine values for all possible indices
        sine_terms_lookup: Precomputed sine terms for all (k1,k2,k3) combinations
        point_results: Output array to store results with shape (N,3)
                       Each row contains [sq, pot, weight]
    """
    # Allocate shared memory for values accessed by all threads in block
    shared_sq = cuda.shared.array(shape=1, dtype=np.int32)
    shared_pot = cuda.shared.array(shape=1, dtype=np.float64)
    
    # Get the current thread's block and thread indices
    point_idx = cuda.blockIdx.x
    thread_idx = cuda.threadIdx.x
    
    # Get number of threads in this block
    block_size = cuda.blockDim.x
    
    # Exit if this block doesn't correspond to a valid point
    if point_idx >= points.shape[0]:
        return
    
    # Get coordinates of the point for this block
    x = points[point_idx, 0]
    y = points[point_idx, 1]
    z = points[point_idx, 2]
    
    # Thread 0 initializes shared memory values
    if thread_idx == 0:
        shared_sq[0] = x*x + y*y + z*z
        shared_pot[0] = 0.0
    
    # Wait for thread 0 to complete initialization
    cuda.syncthreads()
    
    # Each thread will calculate a partial sum of the potential
    local_pot = 0.0
    
    # Divide the workload among threads in the block
    total_iterations = Ns**3
    iterations_per_thread = (total_iterations + block_size - 1) // block_size
    start_idx = thread_idx * iterations_per_thread
    end_idx = min(start_idx + iterations_per_thread, total_iterations)
    
    # Calculate this thread's portion of the potential sum
    for idx in range(start_idx, end_idx):
        # Convert linear index to 3D lattice indices
        k3 = idx % Ns
        temp = idx // Ns
        k2 = temp % Ns
        k1 = temp // Ns
        
        # Skip the origin point in momentum space
        if not (k1 + k2 + k3 == 0):
            # Calculate index for cosine lookup
            cos_idx = (k1*x + k2*y + k3*z) % (3*Ns*Ns)
            r1 = cosf[cos_idx]
            
            # Get the precomputed sine term for these k-indices
            r2_factor = sine_terms_lookup[k1, k2, k3]
            
            # Add this term's contribution to the local sum
            local_pot += r1 * r2_factor
    
    # Add local result to the block's shared sum
    cuda.atomic.add(shared_pot, 0, local_pot)
    
    # Wait for all threads to complete their additions
    cuda.syncthreads()
    
    # Thread 0 finalizes and stores the results
    if thread_idx == 0:
        # Normalize the potential by the lattice volume
        final_pot = shared_pot[0] / Ns**3
        sq = shared_sq[0]
        
        # Store results for this point
        point_results[point_idx, 0] = sq
        point_results[point_idx, 1] = final_pot
        point_results[point_idx, 2] = 1.0  # Weight for later averaging


@cuda.jit
def calculate_x_potential_kernel(points, Ns, cosf,
                                 sine_terms_lookup, point_results):
    """
    CUDA kernel for calculating potentials for points on the x-axis.
    
    This kernel computes the static potential specifically for points along the 
    x-axis (where y=0, z=0). 
    
    Each block processes one spatial point with work distributed across threads.
    
    Args:
        points: Array of x-axis coordinates with shape (N,3), where the y and z
                components are expected to be zero
        Ns: Size of the lattice in each dimension
        cosf: Precomputed cosine values for all possible indices
        sine_terms_lookup: Precomputed sine terms for all (k1,k2,k3) combinations
        point_results: Output array to store results with shape (N,3)
                       Each row contains [sq, pot, weight]
    """
    # Allocate shared memory for values accessed by all threads in block
    shared_sq = cuda.shared.array(shape=1, dtype=np.int32)
    shared_pot = cuda.shared.array(shape=1, dtype=np.float64)
    
    # Get the current thread's block and thread indices
    point_idx = cuda.blockIdx.x
    thread_idx = cuda.threadIdx.x
    
    # Get number of threads in this block
    block_size = cuda.blockDim.x
    
    # Exit if this block doesn't correspond to a valid point
    if point_idx >= points.shape[0]:
        return
    
    # Get x-coordinate of the point for this block (y and z are zero)
    x = points[point_idx, 0]
    
    # Thread 0 initializes shared memory values
    if thread_idx == 0:
        shared_sq[0] = x*x  # Simplified: sq = x**2 for x-axis points
        shared_pot[0] = 0.0
    
    # Wait for thread 0 to complete initialization
    cuda.syncthreads()
    
    # Each thread will calculate a partial sum of the potential
    local_pot = 0.0
    
    # Divide the workload among threads in the block
    total_iterations = Ns**3
    iterations_per_thread = (total_iterations + block_size - 1) // block_size
    start_idx = thread_idx * iterations_per_thread
    end_idx = min(start_idx + iterations_per_thread, total_iterations)
    
    # Calculate this thread's portion of the potential sum
    for idx in range(start_idx, end_idx):
        # Convert linear index to 3D lattice indices
        k3 = idx % Ns
        temp = idx // Ns
        k2 = temp % Ns
        k1 = temp // Ns
        
        # Skip the origin point in momentum space
        if not (k1 + k2 + k3 == 0):
            # Simplified cosine lookup for x-axis points (y=0, z=0)
            cos_idx = (k1*x) % (3*Ns*Ns)
            r1 = cosf[cos_idx]
            
            # Get the precomputed sine term for these k-indices
            r2_factor = sine_terms_lookup[k1, k2, k3]
            
            # Add this term's contribution to the local sum
            local_pot += r1 * r2_factor
    
    # Add local result to the block's shared sum
    cuda.atomic.add(shared_pot, 0, local_pot)
    
    # Wait for all threads to complete their additions
    cuda.syncthreads()
    
    # Thread 0 finalizes and stores the results
    if thread_idx == 0:
        # Normalize the potential by the lattice volume
        final_pot = shared_pot[0] / Ns**3
        sq = shared_sq[0]
        
        # Store results for this point
        point_results[point_idx, 0] = sq
        point_results[point_idx, 1] = final_pot
        point_results[point_idx, 2] = 1.0  # Weight for later averaging


@cuda.jit
def reduce_results_kernel(point_results, pots, weight, num_points):
    """
    CUDA kernel for aggregating individual point calculations into final result arrays.
    
    This kernel takes the individual point results from
    previous calculations and combines them by their squared distance, summing both
    the potential values and weights for subsequent averaging.
    
    The reduction process organizes results by distance rather than by spatial 
    coordinates, allowing the static potential to be properly represented as a
    function of separation distance.
    
    Args:
        point_results: Array of results from individual points with shape (N,3),
                      where each row contains [sq, pot, weight]
        pots: Output array indexed by squared distance, accumulating potential values
        weight: Output array indexed by squared distance, accumulating weights
        num_points: Total number of points to process
    """
    # Get the unique global thread index
    idx = cuda.grid(1)
    
    # Only process if this thread corresponds to a valid point
    if idx < num_points:
        # Extract data for this point
        sq = int(point_results[idx, 0])  # Convert to integer for array indexing
        pot = point_results[idx, 1]      # Potential value for this point
        w = point_results[idx, 2]        # Weight for this point (typically 1.0)
        
        # Atomically add this point's values to the appropriate distance bin
        # Atomic operations are necessary to handle the case where multiple threads
        # attempt to update the same distance bin simultaneously
        cuda.atomic.add(pots, sq, pot)
        cuda.atomic.add(weight, sq, w)


#------------------------------------------------------------------------------
# CPU fallback implementation
#------------------------------------------------------------------------------
def _cpu_impdist(Ns, r2max, improvedAction=True):
    """
    CPU implementation of improved distances calculation.
    Used as fallback when CUDA is not available.
    
    Args:
        Ns: Spatial extension of lattice
        r2max: Maximum squared distance to improve
        improvedAction: Whether to use improved action
        
    Returns:
        List of improved distances
    """
    # Set coefficient based on action type
    cw = 1/3 if improvedAction else 0

    @compile
    def compiled_impdist():
        """ 
        Ported from code by O. Kaczmarek. 
        """
        rimp = []
        kn = 2.0 * np.pi / Ns
        pots = [0.0] * 3 * Ns**2
        weight = [0] * 3 * Ns**2
        
        # Precompute trigonometric values
        cosf = [np.cos(i * kn) for i in range(3 * Ns**2)]
        sinf = [np.sin(i * kn / 2.0) for i in range(Ns)]
        
        # First compute for all x,y,z points up to Ns/4
        for x in range(int(Ns/4) + 1):
            for y in range(int(Ns/4) + 1):
                for z in range(int(Ns/4) + 1):
                    sq = x**2 + y**2 + z**2
                    if sq > r2max:
                        continue
                        
                    pot = 0.0
                    for k1 in range(Ns):
                        for k2 in range(Ns):
                            for k3 in range(Ns):
                                if not (k1+k2+k3) == 0:
                                    r1 = cosf[k1*x + k2*y + k3*z]
                                    r2 = (sinf[k1]**2 + cw*sinf[k1]**4 +
                                          sinf[k2]**2 + cw*sinf[k2]**4 +
                                          sinf[k3]**2 + cw*sinf[k3]**4)
                                    pot += r1 / (4.0 * r2)
                                    
                    pot *= 1.0 / Ns**3
                    pots[sq] += pot
                    weight[sq] += 1
        
        # Then compute for x-axis points from Ns/4+1 to Ns/2
        for x in range(int(Ns/4) + 1, int(Ns/2) + 1):
            sq = x**2
            if sq > r2max:
                continue
                
            pot = 0.0
            for k1 in range(Ns):
                for k2 in range(Ns):
                    for k3 in range(Ns):
                        if not (k1+k2+k3) == 0:
                            r1 = cosf[k1*x]
                            r2 = (sinf[k1]**2 + cw*sinf[k1]**4 +
                                  sinf[k2]**2 + cw*sinf[k2]**4 +
                                  sinf[k3]**2 + cw*sinf[k3]**4)
                            pot += r1 / (4.0 * r2)
                            
            pot *= 1.0 / Ns**3
            pots[sq] += pot
            weight[sq] += 1
        
        # Calculate final improved distances
        for i in range(1, int(r2max) + 1):
            if weight[i] != 0:
                rimp.append(
                    1.0 / (4.0 * np.pi * (pots[i] / weight[i] + 0.22578 / Ns))
                )
                
        return rimp

    return compiled_impdist()

def impdist(Ns, r2max, improvedAction=True):
    """
    GPU-accelerated calculation of tree-level improved distances.
    
    Follows equation (3) of 10.1103/PhysRevD.90.074038.
    Falls back to CPU implementation if CUDA is unavailable.

    Args:
        Ns: Spatial extension of lattice
        r2max: Maximum squared distance to improve
        improvedAction: Whether to use improved action (default: True)

    Returns:
        rimp: List of improved distances

    Raises:
        ValueError: If Ns <= 0 or r2max is too large
    """
    # Input validation
    if not Ns > 0:
        logger.TBError("Need Ns>0")
        
    if r2max > (Ns/2)**2:
        logger.TBError("r2max is too large.")
    
    # Check if CUDA is available
    try:
        if not cuda.is_available():
            logger.warn("CUDA not available, falling back to CPU implementation.")
            return _cpu_impdist(Ns, r2max, improvedAction)
    except:
        logger.warn("CUDA not available, falling back to CPU implementation.")
        return _cpu_impdist(Ns, r2max, improvedAction)

    # Set coefficients based on improved action
    cw = 1/3 if improvedAction else 0
    
    # Prepare data
    kn = 2.0 * np.pi / Ns
    
    # Pre-calculate cosine and sine values
    cosf = np.array([np.cos(i*kn) for i in range(3*Ns**2)], dtype=np.float64)
    sinf = np.array([np.sin(i*kn/2.0) for i in range(Ns)], dtype=np.float64)
    
    # Initialize arrays for potentials and weights
    pots = np.zeros(3*Ns**2, dtype=np.float64)
    weight = np.zeros(3*Ns**2, dtype=np.int32)
    
    # Transfer data to device
    cosf_device = cuda.to_device(cosf)
    sinf_device = cuda.to_device(sinf)
    pots_device = cuda.to_device(pots)
    weight_device = cuda.to_device(weight)
    
    # Get optimal threads per block
    threads_per_block = get_optimal_block_size()
    
    # Create and precompute sine_terms_lookup table to avoid redundant calculations
    sine_terms_lookup = np.zeros((Ns, Ns, Ns), dtype=np.float64)
    sine_terms_device = cuda.to_device(sine_terms_lookup)
    
    # Configure grid for sine terms precomputation
    precompute_threads = threads_per_block
    precompute_blocks = (Ns**3 + precompute_threads - 1) // precompute_threads
    
    # Launch kernel to precompute sine terms
    try:
        compute_sine_terms_kernel[precompute_blocks, precompute_threads](
            sinf_device, cw, Ns, sine_terms_device
        )
    except Exception as e:
        logger.warn('Kernel launch failed. Exception:')
        logger.warn(e)
        logger.warn("CUDA not available, falling back to CPU implementation.")
        return _cpu_impdist(Ns, r2max, improvedAction)

    # Generate xyz points for first calculation phase
    xyz_points = []
    for x in range(int(Ns/4)+1):
        for y in range(int(Ns/4)+1):
            for z in range(int(Ns/4)+1):
                sq = x**2 + y**2 + z**2
                if sq <= r2max:
                    xyz_points.append((x, y, z))
    
    if xyz_points:
        # Convert to numpy array and transfer to device
        xyz_points_array = np.array(xyz_points, dtype=np.int32)
        num_xyz_points = len(xyz_points)
        xyz_points_device = cuda.to_device(xyz_points_array)
        
        # Allocate array for individual point results
        xyz_results = np.zeros((num_xyz_points, 3), dtype=np.float64)  # [sq, pot, weight]
        xyz_results_device = cuda.to_device(xyz_results)
        
        # Configure CUDA grid - one block per point, multiple threads per block
        # Each point is processed by an entire block of threads for parallelism
        blocks_per_grid = num_xyz_points
        
        # Launch optimized kernel - one block per point
        calculate_potential_kernel[blocks_per_grid, threads_per_block](
            xyz_points_device, Ns, cosf_device, sine_terms_device, xyz_results_device
        )
        
        # Reduce individual point results to final arrays
        reduce_blocks = (num_xyz_points + threads_per_block - 1) // threads_per_block
        reduce_results_kernel[reduce_blocks, threads_per_block](
            xyz_results_device, pots_device, weight_device, num_xyz_points
        )
    
    # Generate x-only points for second calculation phase
    x_points = []
    for x in range(int(Ns/4)+1, int(Ns/2)+1):
        sq = x**2
        if sq <= r2max:
            x_points.append((x, 0, 0))
    
    if x_points:
        # Convert to numpy array and transfer to device
        x_points_array = np.array(x_points, dtype=np.int32)
        num_x_points = len(x_points)
        x_points_device = cuda.to_device(x_points_array)
        
        # Allocate array for individual point results
        x_results = np.zeros((num_x_points, 3), dtype=np.float64)  # [sq, pot, weight]
        x_results_device = cuda.to_device(x_results)
        
        # Configure CUDA grid - one block per point, multiple threads per block
        blocks_per_grid = num_x_points
        
        # Launch optimized kernel
        calculate_x_potential_kernel[blocks_per_grid, threads_per_block](
            x_points_device, Ns, cosf_device, sine_terms_device, x_results_device
        )
        
        # Reduce individual point results to final arrays
        reduce_blocks = (num_x_points + threads_per_block - 1) // threads_per_block
        reduce_results_kernel[reduce_blocks, threads_per_block](
            x_results_device, pots_device, weight_device, num_x_points
        )
    
    # Transfer results back to host
    pots_device.copy_to_host(pots)
    weight_device.copy_to_host(weight)
    
    # Calculate improved distances
    rimp = []
    for i in range(1, int(r2max)+1):
        if weight[i] != 0:
            rimp.append(
                1.0 / (4.0 * np.pi * (pots[i] / weight[i] + 0.22578/Ns))
            )
    
    return rimp


# 
# gauge.py                                                               
# 
# D. Clarke
# 
# Tools for reading and writing gauge configurations in Python. To interact with binary configurations, we can use
# Python's built-in struct module. More info on that: https://docs.python.org/3.7/library/struct.html. NERSC reader inspired
# by https://github.com/mdipierro/qcdutils. ILDG reader from https://github.com/nftqcd/nthmc/blob/master/lib/fieldio.py
#

import numpy as np
import struct, os
import latqcdtools.base.logger as logger
from latqcdtools.base.speedify import numbaON,DEFAULTTHREADS
from latqcdtools.math.math import rel_check
from latqcdtools.base.check import checkType
from latqcdtools.interfaces.lime import LIMEMAGIC,trimNull,xmlFind,scidacChecksum
numbaON()
from latqcdtools.math.SU3 import SU3
from latqcdtools.physics.gauge import gaugeField


class confReader:

    """ 
    Base class for reading configurations. 
    """

    def __init__(self, Ns, Nt, nproc=None):
        """ 
        Initialize a confReader object.

        Args:
            Ns (int): Spatial extension of lattice. 
            Nt (int): Euclidian time extension of lattice. 
            nrows (int): Number of saved rows for compression. 
        """        
        self.Ns = Ns
        self.Nt = Nt
        self.Nd = 4
        self.Nc = 3
        checkType('int',Ns=Ns)
        checkType('int',Nt=Nt)
        if nproc==None:
            self.nproc=min(self.Nt,DEFAULTTHREADS)
        else:
            checkType('int',nproc=nproc)
            self.nproc=nproc
        self.offset = None     # How many bytes is the header?
        self.endianness = '>'  # Big endian by default
        self.precision = 'f'   # Single precision by default
        self.file = None       # File handle
        self.linkTrace = None  # <tr U>
        self.plaquette = None  # <tr U^[]>
        self.nrows = None      # number of rows
        self.gauge = gaugeField(self.Ns,self.Nt,self.nproc)


    def unpack(self, data) -> SU3:
        """ 
        Unpack a string of bytes from file into an SU3 object. 
        """
        numData = int(len(data)/self.getByteSize())
        linkTuple = struct.unpack( self.endianness + str(numData) + self.precision, data )
        link = SU3()
        i = 0
        for j in range(self.rows):
            for k in range(3):
                link[j,k] = complex( linkTuple[i], linkTuple[i+1] )
                i += 2
        if self.rows==2:
            link.su3unitarize()
        return link


    def getByteSize(self):
        if self.precision == 'f':
            return 4
        elif self.precision == 'd':
            return 8
        else:
            logger.TBRaise('Unknown precision',self.precision,'(expected f or d)')


    def checkLatDims(self, Nx, Ny, Nz, Nt):
        if Nx != self.Ns:
            logger.TBRaise('Read Nx =',Nx,'. Expected ',self.Ns)
        if Ny != self.Ns:
            logger.TBRaise('Read Ny =',Ny,'. Expected ',self.Ns)
        if Nz != self.Ns:
            logger.TBRaise('Read Nz =',Nz,'. Expected ',self.Ns)
        if Nt != self.Nt:
            logger.TBRaise('Read Nt =',Nt,'. Expected ',self.Nt)



class NERSCReader(confReader):

    def readHeader(self,fileName):

        """ 
        Extract metadata from the header. Do a few checks to make sure the read-in went smoothly. This method
        assumes all NERSC headers have the minimal entries
            BEGIN_HEADER
            DATATYPE
            DIMENSION_1
            DIMENSION_2
            DIMENSION_3
            DIMENSION_4
            FLOATING_POINT
            END_HEADER
        """

        self.file = open(fileName,'rb')

        # Extract offset.
        header = self.file.read()
        minOffset = 78 # Minimal number of characters for the entries above, not including what comes on RHS.
        self.offset = header.find(b'END_HEADER\n')+11
        if self.offset < minOffset:
            logger.TBRaise(fileName,'not in NERSC format.')

        # Convert header to metaData dictionary.
        entries = header[:self.offset-1].split(b'\n')[1:-1]
        metaData = {}
        for entry in entries:
            LHS = entry.split(b'=')[0].strip()
            RHS = entry.split(b'=')[1].strip()
            metaData[LHS] = RHS

        # Check that the header was read correctly.
        Nx = int(metaData[b'DIMENSION_1'])
        Ny = int(metaData[b'DIMENSION_2'])
        Nz = int(metaData[b'DIMENSION_3'])
        Nt = int(metaData[b'DIMENSION_4'])
        self.checkLatDims(Nx, Ny, Nz, Nt)

        # Extract endianness
        if metaData[b'FLOATING_POINT'].endswith(b'SMALL'):
            self.endianness = '<'
        elif metaData[b'FLOATING_POINT'].endswith(b'BIG'):
            self.endianness = '>'
        else:
            logger.TBRaise('Unrecognized endianness.')

        # Extract precision
        if metaData[b'FLOATING_POINT'].startswith(b'IEEE32'):
            self.precision = 'f'
        elif metaData[b'FLOATING_POINT'].startswith(b'IEEE64'):
            self.precision = 'd'
        else:
            logger.TBRaise('Unrecognized precision.')

        # Extract number of rows
        if metaData[b'DATATYPE'].endswith(b'3x3'):
            self.rows = 3
        else:
            self.rows = 2

        # Extract trace of average link
        try:
            self.linkTrace = float( metaData[b'LINK_TRACE'] )
        except:
            pass

        # Extract average plaquette
        try:
            self.plaquette = float( metaData[b'PLAQUETTE'] )
        except:
            pass


    def readConf(self,fileName):
        """ 
        Read in the configuration. Check what you find in the binary against its metadata. 
        """

        self.readHeader(fileName)

        bytesPerLink = 3*self.rows*2*self.getByteSize()

        for t in range(self.Nt):
            for z in range(self.Ns):
                for y in range(self.Ns):
                    for x in range(self.Ns):
                        for mu in range(self.Nd):
                            byteIndex = self.offset + bytesPerLink*( mu + 4*( x + self.Ns*( y + self.Ns*( z + self.Ns*t ) ) ) )
                            self.file.seek(byteIndex)
                            data = self.file.read(bytesPerLink)
                            link = self.unpack(data)
                            self.gauge.setLink(link,x,y,z,t,mu)
        self.file.close()
        logger.details('Loaded conf into gaugeField.')

        # Check <tr U>
        if self.linkTrace is not None:
            linkTrace = self.gauge.getLinkTrace()
            if not rel_check(linkTrace, self.linkTrace):
                logger.TBRaise('<tr U> is wrong. Compare:',linkTrace,'with',self.linkTrace)
            logger.details('Configuration',fileName,'has correct <tr U>.')

        # Check <tr U^[]>
        if self.plaquette is not None:
            plaq = self.gauge.getPlaquette()
            if not rel_check(plaq, self.plaquette):
                logger.TBRaise('<tr U^[]> is wrong. Compare:',plaq,'with',self.plaquette)
            logger.details('Configuration', fileName, 'has correct <tr U_plaq>.')

        return self.gauge



class ILDGReader(confReader):

    def readConf(self,fileName):
        """ 
        Read in the configuration. Check the checksums if they exist. Ignore spin. 
        """

        self.file       = open(fileName,'rb')
        self.nrows      = 3
        self.offset     = 144
        self.endianness = '>'

        latsize = -1
        bytesPerLink = -1
        latdatacount = -1
        latdims = []
        latnc = -1
        latprec = -1
        lattype = b'Unknown'

        while True:

            header = self.file.read(self.offset)
            if not header:
                break

            magi, vers, mbeg_end_res, size, type = struct.unpack('>ihHq128s',header)
            if magi!=LIMEMAGIC:
                logger.TBRaise(f'lime magic number does not match, got {magi}')
            type = trimNull(type)

            if type==b'scidac-binary-data' or type==b'ildg-binary-data':
                lattype = type
                latsize = size
                latdata = self.file.read(size)
                next = 7-(size+7)%8
            else:
                data = self.file.read(size)
                data = trimNull(data)
                if type==b'scidac-private-file-xml':
                    spacetime = int(xmlFind(data,'spacetime'))
                    latdims = [int(x) for x in xmlFind(data,'dims').strip().split()]
                    if len(latdims)!=spacetime:
                        logger.TBRaise(f'got spacetime {spacetime} but dims {latdims}')
                elif type==b'scidac-file-xml':
                    logger.info(f'file metadata: {data}')
                elif type==b'scidac-private-record-xml':
                    logger.info('date: ',xmlFind(data,'date'))
                    logger.info('recordtype: ',xmlFind(data,'recordtype'))
                    logger.info('datatype: ',xmlFind(data,'datatype'))
                    precision = xmlFind(data,'precision').lower()
                    if precision==b'f':
                        latprec = 8    # complex float
                    elif precision==b'd':
                        latprec = 16   # complex double
                    else:
                        logger.TBRaise(f'unknown precision {precision}')
                    latnc = int(xmlFind(data,'colors'))
                    bytesPerLink = int(xmlFind(data,'typesize'))
                    latdatacount = int(xmlFind(data,'datacount'))
                elif type==b'scidac-record-xml':
                    logger.info(f'record metadata: {data}')
                elif type==b'scidac-checksum':
                    latsuma = int(xmlFind(data,'suma'),16)
                    latsumb = int(xmlFind(data,'sumb'),16)
                    pass
                elif type==b'ildg-format':
                    field = xmlFind(data,'field')
                    if field!=b'su3gauge':
                        logger.TBRaise(f'unsupported ildg field type {field}')
                    precision = int(xmlFind(data,'precision'))
                    if precision==32:
                        latprec = 8    # complex float
                    elif precision==64:
                        latprec = 16   # complex double
                    else:
                        logger.TBRaise(f'unknown precision {precision}')
                    lx = int(xmlFind(data,'lx'))
                    ly = int(xmlFind(data,'ly'))
                    lz = int(xmlFind(data,'lz'))
                    lt = int(xmlFind(data,'lt'))
                    latdims = [lx,ly,lz,lt]
                    latnc = self.Nc
                    bytesPerLink = latnc*latnc*latprec
                    latdatacount = self.Nd
                else:
                    logger.info(f'unused type: {type}  data: {data}')
                next = 7-(size+7)%8
            self.file.seek(next,os.SEEK_CUR)
        self.file.close()
        ndim = len(latdims)

        # Carry out some checks on the read-in data.
        if latsize<=0 or bytesPerLink<=0 or latdatacount<=0 or ndim==0 or latnc<0 or latprec<0 or lattype==b'Unknown':
            logger.TBRaise(f'unsupported file: {fileName}')
        if latprec==8:
            self.precision = 'f'
        elif latprec==16:
            self.precision = 'd'
        else:
            logger.TBRaise(f'unknown complex precision {latprec}')
        if latnc != self.Nc:
            logger.TBRaise(f'unsupported number of colors: {latnc}')
        if latdatacount != ndim:
            logger.TBRaise(f'There should be one link per direction, but header suggests {latdatacount} links.')
        self.checkLatDims(*latdims)
        if ndim!=self.Nd:
            logger.TBRaise(f'unsupported number of dimensions: {ndim}.')
        vol = self.Ns**3 * self.Nt 
        if latsize != vol*self.Nd*bytesPerLink:
            logger.TBRaise(f'incorrect lattice size, expect {vol*ndim*bytesPerLink}, but got {latsize}')
        if bytesPerLink != latprec*latnc*latnc:
            logger.TBRaise(f'incorrect bytes per link, expected {latprec*latnc*latnc}, but got {bytesPerLink}')
        if not (lattype==b'scidac-binary-data' or lattype==b'ildg-binary-data'):
            logger.TBRaise(f'unknown lattice format: {lattype}')

        # Check checksums
        if lattype==b'scidac-binary-data':
            if latsuma is None or latsumb is None:
                logger.TBRaise(f'No SciDAC checksum.')
            suma,sumb = scidacChecksum(latdata, vol, ndim*bytesPerLink)
            if suma!=latsuma or sumb!=latsumb:
                logger.TBRaise(f'Checksum error: expected {latsuma:x} {latsumb:x}, computed {suma:x} {sumb:x}')
            logger.info('SciDAC checksum OK.')

        # Convert from bytes to numpy array.
        lat = np.frombuffer(latdata,dtype='>c'+str(latprec),count=vol*ndim*latnc*latnc)
        lat = np.reshape(lat,latdims[::-1]+[ndim,latnc,latnc])
        
        # Load into gaugeField.
        for t in range(self.Nt):
            for z in range(self.Ns):
                for y in range(self.Ns):
                    for x in range(self.Ns):
                        for mu in range(self.Nd):
                            link = lat[t,z,y,x,mu]
                            self.gauge.setLink(link,x,y,z,t,mu)
        logger.details('Loaded conf into gaugeField.')

        return self.gauge 



#
# referenceScales.py
#
# A collection of scales computed as a function of beta. 
#

import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.math.polynomials import Polynomial, Rational
from latqcdtools.physics.runningCoupling import beta_func, b0, b1


CHECKBETARANGE = True


# Most recent update of scales in physical units
CY_phys  = { 
            'fk' : 2019,
            'r1' : 2010,
            'r0' : 2014,
            't0' : 2015 
            }


# Most recent update of scale parameterizations
CY_param = { 
            'fk' : 2021,
            'r1' : 2021,
            'r0' : 2017,
            't0' : 2015 
            }


def ignoreBetaRange():
    """ 
    Turn off the beta range warnings. 
    """
    global CHECKBETARANGE
    CHECKBETARANGE = False
    logger.warn('Squelching beta range warnings.')


def _betaRangeWarn(beta, beta_range):
    """ 
    Many of these ansätze a(beta) have coefficients that were determined by performing a fit within a certain
    beta range. This warning flashes whenever you are using information outside of that range, where the ansatz
    is less likely be to reliable.

    Args:
        beta (float) or numpy array
        beta_range (array-like): min and max beta of range, in that order 
    """
    global CHECKBETARANGE
    if CHECKBETARANGE:
        if isinstance(beta , (float,np.floating)) :
            if beta < beta_range[0] or beta > beta_range[1]:
                logger.warn(f"beta out of fit range [{beta_range[0]},{beta_range[1]}]",frame=3)
        else:
            if np.min(beta) < beta_range[0] or np.max(beta) > beta_range[1]:
                logger.warn(f"beta out of fit range [{beta_range[0]},{beta_range[1]}]",frame=3)


# -------------------------------------------------------------------------------------------------------- FITTING FORMS


def fit_2014Eos_eqB2(beta,c0,c2,d2):
    return Rational([0,c0,0,c2*10/beta],[1,0,d2*10/beta])(beta_func(beta))


def fit_tayloraLambda(beta,a,b,c):
    return Polynomial([0,a,0,b,0,c])(beta_func(beta))


# ----------------------------------------------------------------------------------------------- 2+1 FLAVOR HISQ SCALES


def allton_type_ansatz(beta, c0, c2, d2):
    return (c0*beta_func(beta) + c2*(10/beta)*beta_func(beta)**3) / \
           (1 + d2*(10/beta)*beta_func(beta)**2)


# ===================================================== f_K scales


def a_times_fk(beta, year):
    """ 
    Get a*f_k(beta).

    Args:
        beta (float)
        year (int/str): year that parameterization was determined 

    Returns:
        float: a*f_k 
    """
    # https://arxiv.org/pdf/2107.10011.pdf, 10.1103/PhysRevD.104.074512
    if str(year) == "2021":
        beta_range = [6.175, 7.220]
        _betaRangeWarn(beta, beta_range)
        c0fk = 7.486
        c2fk = 41935.0
        d2fk = 3273.0

    # 10.1103/PhysRevD.100.094510
    elif str(year) == "2014":
        beta_range = [6., 7.373]
        _betaRangeWarn(beta, beta_range)
        c0fk = 7.49415
        c2fk = 46049.0
        d2fk = 3671.0

    # https://arxiv.org/pdf/1111.1710 , 10.1103/PhysRevD.85.054503 
    elif str(year) == "2012":
        beta_range = [6., 6.8]
        _betaRangeWarn(beta, beta_range)
        c0fk = 7.65667
        c2fk = 32911.0
        d2fk = 2388.0

    else:
        logger.TBRaise("No fit parameters for ", str(year))

    return allton_type_ansatz(beta, c0fk, c2fk, d2fk)


# ====================================================== r1 scales


def a_div_r1(beta, year):
    """ 
    Get a/r_1(beta).

    Args:
        beta (float)
        year (int/str): year that parameterization was determined 

    Returns:
        float: a/r_1
    """
    # https://arxiv.org/pdf/2107.10011.pdf. Note that in this study, they say they update
    # numbers from https://arxiv.org/pdf/1407.6387.pdf making changes that seem unrelated
    # to the beta range, while providing no new beta range. I therefore assume the beta
    # range is the same as Table IV of https://arxiv.org/pdf/1407.6387.pdf
    if str(year) == "2021":
        beta_range = [5.9, 7.825]
        _betaRangeWarn(beta, beta_range)
        c0 = 43.16
        c2 = 339472
        d2 = 5452.0
    # https://arxiv.org/pdf/1710.05024.pdf
    elif str(year) == "2018":
        beta_range = [7.030, 8.4]
        _betaRangeWarn(beta, beta_range)
        c0 = 43.12
        c2 = 347008
        d2 = 5584
    # https://arxiv.org/pdf/1407.6387.pdf
    elif str(year) == "2014":
        beta_range = [5.9, 7.825]
        _betaRangeWarn(beta, beta_range)
        c0 = 43.1
        c2 = 343236.0
        d2 = 5514.0
    # https://arxiv.org/pdf/1111.1710.pdf. It should be noted that this one has its LCP
    # tuned at ms/ml=20.
    elif str(year) == "2012":
        beta_range = [5.9, 7.28]
        _betaRangeWarn(beta, beta_range)
        c0 = 44.06
        c2 = 272102.0
        d2 = 4281.0
    else:
        logger.TBRaise("No fit parameters for year", str(year))
    return allton_type_ansatz(beta, c0, c2, d2)


# ================================================= strange quark mass: line of constant physics


# fit taken from 1407.6387v2
def r1_times_ms_2014(beta):
    nf   = 3
    mRGI = 0.2609
    m1   = 35600
    m2   = -21760
    m3   = 2.67*10**7
    dm1  = 2420
    num  = 1 +  m1*(10.0/beta)*beta_func(beta)**2+m2*(10.0/beta)**2*beta_func(beta)**2 + m3*(10.0/beta)*beta_func(beta)**4
    den  = 1 + dm1*(10.0/beta)*beta_func(beta)**2
    return (20*b0(nf)/beta)**(4/9)*mRGI*num/den


def a_times_ms_2014(beta):
    return r1_times_ms_2014(beta) * a_div_r1(beta, "2014")


# ------------------------------------------------------------------------------------------------------ QUENCHED SCALES


def wuppertal_type_ansatz(beta,c1,c2,c3,c4):
    Nf = 0
    B0 = b0(Nf)/(4*np.pi)**2
    B1 = b1(Nf)/(4*np.pi)**4
    return np.exp( (beta/(12*B0)+B1/(2.*B0**2)*np.log(6*B0/beta))
                   * (1+c1/beta+c2/beta**2)/(1+c3/beta+c4/beta**2) )


# ===================================================== r0 scales


def r0_div_a(beta,year):
    """ 
    Get r0/a(beta).

    Args:
        beta (float)
        year (int/str): year that parameterization was determined 

    Returns:
        float: r0/a 
    """
    # https://arxiv.org/abs/1709.07612
    if str(year) == "2017":
        c1 = -8.9664
        c2 = 19.21
        c3 = -5.25217
        c4 = 0.606828
    # https://arxiv.org/abs/1503.05652
    elif str(year) == "2015":
        beta_range = [5.7, 6.92]
        _betaRangeWarn(beta, beta_range)
        c1 = -8.17273
        c2 = 14.9600
        c3 = -3.95983
        c4 = -5.30334
    # 10.1103/PhysRevD.85.054503
    elif str(year) == "2012":
        c0 = 32.83
        c2 = 81127
        d2 = 1778
        return 1/allton_type_ansatz(beta,c0,c2,d2)
    else:
        logger.TBRaise("No fit parameters for year", str(year))
    return wuppertal_type_ansatz(beta,c1,c2,c3,c4)


# ===================================================== t0 scales


# Based on https://arxiv.org/pdf/1503.05652.pdf
# Latest update at 2017/01/11 by Lukas Mazur
def sqrtt0_div_a(beta):
    """ 
    Get sqrt(t0/a)(beta)

    Args:
        beta (float)

    Returns:
        float: sqrt(t0/a) 
    """
    c1 = -9.945
    c2 = 24.191
    c3 = -5.334
    c4 = 1.452
    return wuppertal_type_ansatz(beta,c1,c2,c3,c4)



#
# printErrorBars.py
#
# H. Sandmeyer
#
# Methods for printing measurements with error bars as they typically appear in lattice publications. The default is
# 2 significant digits on an error bar in parentheses: measurement = X.XXXX(YY)
#

import numpy as np 
import latqcdtools.base.logger as logger
from latqcdtools.base.check import checkType


def getValuesFromErrStr(errStr):
    """ 
    Convert A string of the form XX.XX(YY) into a float mean and error bar. Scientific notation not yet supported.
    Single digit errors not yet supported.

    Args:
        errStr (str): string of the form XX.XXXX(YY).

    Returns:
        float, float: mean, error. 
    """
    checkType(str,errStr=errStr)
    try:
        meanStr  = errStr.split('(')[0]
        mean     = float(meanStr)
        ebstr    = errStr.split('(')[1][:-1]
        err      = float(ebstr)
        if not '.' in errStr:
            return mean, err
        dot_index = errStr.find('.')
        parenthesis_index = errStr.find('(')
        N = parenthesis_index - dot_index - 1
        if get_exp(err)==0 and '.' in ebstr:
            N -= 1
        err *= pow(10,-N)
        return mean, err        
    except Exception as e:
        logger.TBRaise('Expected string of form XX.XXXX(YY). Got',errStr,exception=e)


def get_exp(param):
    """ 
    Get the exponent of a number to base 10. 
    """
    result = np.floor( np.log(abs(param))/np.log(10) )
    if np.isnan(result):
        return np.nan
    else:
        return int(result)


def get_err_str(param, param_err, numb_err_dig=2, rounding='conservative') -> str:
    """
    Get the string of a number + error, e.g. 1.234567+-0.324456 --> 12.34(33) (numb_err_dig = 2). 

    Args:
        param (float): _description_
        param_err (float): _description_
        numb_err_dig (int, optional): How many significant digits for your error? Defaults to 2.
        rounding (str, optional): The strategy for rounding the last significant digit of the error.
            The 'canonical' strategy rounds how one learns in grade school, i.e. 44 gets rounded to 40.
            The 'conservative' strategy always rounds up, i.e. 44 gets rounded to 50. Defaults to
            'conservative'. 

    Returns:
        str: Error string. 
    """
    checkType('real',param=param)
    checkType('real',param_err=param_err)
    checkType('int',numb_err_dig=numb_err_dig)
    checkType(str,rounding=rounding)

    if param_err<=0:
        logger.details('Encountered non-positive error',param_err)
        return param

    if numb_err_dig < 1:
        logger.TBRaise("Number of error digits has to be larger than 0!")
    if param < 0:
        param = -param
        sign = -1
    else:
        sign = 1

    relnum = get_exp(param_err)
    if np.isnan(relnum):
        return "%.12e" % param

    # index for rounding the error and the parameter
    roundidx = -relnum + numb_err_dig - 1
    paramtmp = param
    param = np.round(param, roundidx)

    # exponent of the actual parameter
    if param == 0.0:
        numdig=get_exp(param_err)
    else:
        numdig = get_exp(param)

    # Number of digits before the dot
    if numdig < 0:
        numdig = 0

    # floor does not support a second index -> we have to multiply and then divide
    param_err *= pow(10, roundidx)
    if rounding=='conservative':
        param_err = np.ceil(param_err) * pow(10, -roundidx)
    elif rounding=='canonical':
        param_err = np.round(param_err) * pow(10, -roundidx)
    else:
        logger.TBRaise('Unknown rounding strategy',rounding)

    # as the exponent might have changed through rounding, we have to recalculate it
    relnum = get_exp(param_err)
    roundidx = -relnum + numb_err_dig - 1
    param = np.round(paramtmp, roundidx)

    # Strings that are shortened later on
    err_str = "%.12lf" % param_err
    param_str = "%.12lf" % param

    # if the error is larger than or equal to 1
    if relnum >= 0:
        # if we need more digits to express the full error than given by numb_err_digits
        if numb_err_dig <= relnum + 1:
            err_str = err_str[:relnum + 1]
            # cut parameter before the dot
            param_str = param_str[:numdig + 1]
        else:
            # if we have to print more digits than relnum
            err_str = err_str[:numb_err_dig + 1]
            # also print after the dot with the relevant number of digits
            param_str = param_str[:numdig + 1 + (numb_err_dig - relnum)]
    else:
        # we don't need the part before the dot here
        err_str = err_str[-relnum + 1:-relnum + numb_err_dig + 1]
        param_str = param_str[:numdig - relnum + numb_err_dig + 1]

    if sign == -1:
        return "-%s(%s)" % (param_str, err_str)
    else:
        return "%s(%s)" % (param_str, err_str)


def get_err_str_exp(param, param_err, exp, numb_err_dig=1, multicon="x") -> str:
    """ 
    Express the number with a exponent. The multiplication icon can be changed. 
    """
    if exp == 0:
        return get_err_str(param, param_err, numb_err_dig)
    param /= pow(10, exp)
    param_err /= pow(10, exp)
    return "%s%s10^%i" % (get_err_str(param, param_err, numb_err_dig), multicon, exp)


def get_err_str_auto(param, param_err, numb_err_dig=1, mulicon="x") -> str:
    """ 
    Automatically express the number with an exponent. The multiplication icon can be changed. 
    """
    exp = get_exp(param)
    exp_err = get_exp(param_err)
    if exp_err > exp:
        exp = exp_err
    if 4 > exp > -4:
        return get_err_str(param, param_err, numb_err_dig)
    return get_err_str_exp(param, param_err, exp, numb_err_dig, mulicon)


def get_err_str_exp_tex(param, param_err, exp, numb_err_dig=1) -> str:
    """ 
    Express the number with an exponent in LaTeX. 
    """
    return "$%s$" % get_err_str_exp(param, param_err, exp, numb_err_dig, "\\cdot ")


def get_err_str_auto_tex(param, param_err, numb_err_dig=1) -> str:
    """ 
    Automatically express the number with an exponent in LaTeX. 
    """
    return "$%s$" % get_err_str_auto(param, param_err, numb_err_dig, "\\cdot ")



# 
# gauge.py                                                               
# 
# D. Clarke
# 
# A gauge field class. We try to implement various functions as static methods so that we can compile
# them with numba to speed them up. We assume periodic boundary conditions.
# 

import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.math.SU3 import SU3
from latqcdtools.math.math import id
from latqcdtools.base.speedify import compile, parallel_reduce, parallel_function_eval
from latqcdtools.base.check import checkType


@compile
def respectBCs(x,y,z,t,Ns,Nt):
    """ 
    Assume periodic BCs. This also seems to work for negative numbers, e.g. -1 % 8 = 7. 
    """
    return x%Ns, y%Ns, z%Ns, t%Nt


@compile
def stepUp(x,y,z,t,mu,Ns,Nt):
    X, Y, Z, T = x, y, z, t
    if mu==0:
        X += 1
    elif mu==1:
        Y += 1
    elif mu==2:
        Z += 1
    elif mu==3:
        T += 1
    return respectBCs(X, Y, Z, T, Ns, Nt)


@compile
def stepDown(x,y,z,t,mu,Ns,Nt):
    X, Y, Z, T = x, y, z, t
    if mu==0:
        X -= 1
    elif mu==1:
        Y -= 1
    elif mu==2:
        Z -= 1
    elif mu==3:
        T -= 1
    return respectBCs(X, Y, Z, T, Ns, Nt)


@compile
def ReTrABCD(A,B,C,D):
    return np.trace( A @ B @ C @ D ).real


class gaugeField:

    """ 
    A class for SU3 gauge fields. 
    """

    def __init__(self, Ns, Nt, nproc=1):
        """ 
        Initialize an SU3 gaugeField.

        Args:
            Ns (int): Spatial extension of lattice. 
            Nt (int): Euclidian time extension of lattice. 
            nproc (int, 1): Number of proccesors for parallelization. Defaults to 1.
        """
        logger.details('Initialize gaugeField object...')
        checkType('int',Ns=Ns)
        checkType('int',Nt=Nt)
        checkType('int',nproc=nproc)
        self.Ns    = Ns
        self.Nt    = Nt
        self.nproc = nproc
        self.field = np.zeros(shape=(self.Nt,self.Ns,self.Ns,self.Ns,4,3,3),dtype=complex)


    def __repr__(self) -> str:
        return "gaugeField"


    def getLink(self,x,y,z,t,mu) -> SU3:
        """ 
        Link accessor. 
        """
        X, Y, Z, T = respectBCs(x,y,z,t,self.Ns,self.Nt)
        return SU3(self.field[T][Z][Y][X][mu])


    def setLink(self, other, x, y, z, t, mu):
        """ 
        Link setter. 
        """
        X, Y, Z, T = respectBCs(x,y,z,t,self.Ns,self.Nt)
        self.field[T][Z][Y][X][mu] = other


    def getLocalPlaquette(self,x,y,z,t,mu,nu):
        """ 
        Get Re tr U_{mu,nu}^[](x,y,z,t). 
        """
        U_mu_x     = SU3()
        U_nu_xpamu = SU3()
        U_mu_xpanu = SU3()
        U_nu_x     = SU3()
        U_mu_x.setToMatrix( self.getLink(x,y,z,t,mu) )
        X, Y, Z, T = stepUp(x,y,z,t,mu,self.Ns,self.Nt)
        U_nu_xpamu.setToMatrix( self.getLink(X,Y,Z,T,nu) )
        X, Y, Z, T = stepUp(x,y,z,t,nu,self.Ns,self.Nt)
        U_mu_xpanu.setToMatrix( self.getLink(X,Y,Z,T,mu) )
        U_nu_x.setToMatrix( self.getLink(x,y,z,t,nu) )
        return ReTrABCD(U_mu_x,U_nu_xpamu,U_mu_xpanu.dagger(),U_nu_x.dagger())


    def getPlaquette(self):
        """ 
        Calculate <Re tr U_{mu,nu}^[]>. 
        """
        plaq = parallel_reduce(self._plaq_contrib,range(self.Nt),nproc=self.nproc)
        return plaq/(self.Ns**3*self.Nt*18)


    def _plaq_contrib(self,t):
        contrib = 0
        for z in range(self.Ns):
            for y in range(self.Ns):
                for x in range(self.Ns):
                    for nu in range(1,4):
                        for mu in range(nu):
                            contrib += self.getLocalPlaquette(x,y,z,t,mu,nu)
        return contrib


    def getLinkTrace(self):
        """ 
        Calculate <tr U>. 
        """
        linkTrace = parallel_reduce(self._linkTrace_contrib,range(self.Nt),nproc=self.nproc)
        return linkTrace/(self.Ns**3*self.Nt*4*3)


    def _linkTrace_contrib(self,t):
        contrib = 0
        for z in range(self.Ns):
            for y in range(self.Ns):
                for x in range(self.Ns):
                    for mu in range(4):
                        contrib += self.getLink(x,y,z,t,mu).trace().real
        return contrib


    def makeCold(self):
        """ 
        Cold start, i.e. every link is set to the identity. 
        """
        for t in range(self.Nt):
            for z in range(self.Ns):
                for y in range(self.Ns):
                    for x in range(self.Ns):
                        for mu in range(4):
                            self.setLink(id(3),x,y,z,t,mu)


    def checkSU3(self):
        parallel_function_eval(self._checkSU3_timeslice, range(self.Nt), nproc=self.nproc)
        logger.info('All links are SU(3).')


    def _checkSU3_timeslice(self,t):
        for z in range(self.Ns):
            for y in range(self.Ns):
                for x in range(self.Ns):
                    for mu in range(4):
                        if not self.getLink(x,y,z,t,mu).isSUN():
                            logger.TBRaise(f'Link at ({x},{y},{z},{t}) in direction {mu} is not SU(3).')



# 
# logger.py                                                               
# 
# H. Sandmeyer, D. Clarke 
# 
# Methods for logging and output.
# 

import sys, inspect, datetime, logging
from colorama import Fore


_PASS         = Fore.GREEN
_WARNING      = Fore.YELLOW 
_FAIL         = Fore.RED 
_ENDC         = '\033[0m'
RECORDLOG     = False
CURRENT_LEVEL = 4


class ToolboxException(Exception): pass

# These modules are not allowed to use the logger
_BANLIST = ['matplotlib']


_log_levels = {
    'ALL'      : 0,
    'DEBUG'    : 1,
    'DETAILS'  : 2,
    'PROGRESS' : 3,
    'INFO'     : 4,
    'WARN'     : 5,
    'NONE'     : 6
    }


def createLogFile(filename="Toolbox.log"):
    """ 
    Have output sent also to a log file filename. If this file already exists, it will get deleted. We use the
    logging module because it knows how to handle multiple processes writing to the same file. 
    """
    global RECORDLOG
    RECORDLOG = True
    # I recommend against changing the log level here. Other modules, such as Numba, also use this logger, and
    # hence you can get spammed if you set it to DEBUG. Similarly, keep the format the same, which otherwise will
    # print additional, unwanted strings at the beginning of each line.
    logging.basicConfig(filename=filename, encoding='utf-8', level=logging.INFO, format='%(message)s', filemode='w')
    for module in _BANLIST:
        logging.getLogger(module).setLevel(logging.CRITICAL)
    info('Created log file',filename)


def _log(outString):
    global RECORDLOG
    if RECORDLOG:
        logging.info(outString[:-1])


def _getCallerName(frame):
    """ 
    Gets the name of the function that calls the present function. 
    """
    # The way the frame works: Each nested function is labelled by the first index. 0 is the current function, i.e.
    # getCallerName. 1 is the function that called this function, and so on. Second index = 3 retrieves the name.
    currframe  = inspect.currentframe()
    callframe  = inspect.getouterframes(currframe, 2)
    methodName = str(callframe[frame][3])+': '
    # If the main script is calling it, we don't need a name.
    if methodName=='<module>: ':
        methodName='' 
    # If the method was called from within a class, get the class name too.
    call_locals = currframe.f_back.f_back.f_locals
    try:
        className = str(call_locals['self'])+'.'
    except KeyError:
        try:
            className = str(call_locals['locals_dict']['self'])+'.'
        except KeyError:
            className = ''
    return className+methodName


def _getTimeStamp():
    """ 
    Get HH:MM:SS 
    """
    return ' ['+datetime.datetime.now().strftime("%H:%M:%S")+']'


# ----------------------------------------------------------------------------------------------- DEPENDENT ON LOG LEVEL


def set_log_level(level):
    global CURRENT_LEVEL
    CURRENT_LEVEL = _log_levels[level]


def debug(*args,frame=2):
    global CURRENT_LEVEL
    if CURRENT_LEVEL <= 1:
        args   = [str(s) for s in args]
        output = _getTimeStamp()+' DEBUG: '+_getCallerName(frame)+(' '.join(args))
        print(output)
        _log(output+'\n')


def details(*args):
    global CURRENT_LEVEL
    if CURRENT_LEVEL <= 2:
        args   = [str(s) for s in args]
        output = _getTimeStamp()+' DETAILS: '+(' '.join(args))
        print(output)
        _log(output+'\n')


def progress(*args):
    global CURRENT_LEVEL
    if CURRENT_LEVEL <= 3:
        args   = [str(s) for s in args]
        output = _getTimeStamp()+' PROGRESS: '+(' '.join(args))
        print(output)
        _log(output+'\n')


def info(*args):
    global CURRENT_LEVEL
    if CURRENT_LEVEL <= 4:
        args   = [str(s) for s in args]
        output = _getTimeStamp()+' INFO: '+(' '.join(args))
        print(output)
        _log(output+'\n')


def warn(*args,frame=2):
    global CURRENT_LEVEL
    if CURRENT_LEVEL <= 5:
        args    = [str(s) for s in args]
        output  = _getTimeStamp()+' WARNING: '+_getCallerName(frame)+(' '.join(args))
        outputc = _getTimeStamp()+_WARNING+' WARNING: '+_getCallerName(frame)+(' '.join(args))+_ENDC
        print(outputc)
        _log(output+'\n')


# --------------------------------------------------------------------------------------------- INDEPENDENT OF LOG LEVEL


def TBFail(*args):
    args    = [str(s) for s in args]
    output  = _getTimeStamp()+' FAIL: '+(' '.join(args))
    outputc = _getTimeStamp()+_FAIL+' FAIL: '+(' '.join(args))+_ENDC
    print(outputc)
    _log(output + '\n')


def TBError(*args,frame=2):
    """
    Print error message and exit with -1.

    Args:
        frame (int, optional): Controls the name of the caller. Defaults to method that called TBError.
    """
    args    = [str(s) for s in args]
    output  = _getTimeStamp()+' ERROR: '+_getCallerName(frame)+(' '.join(args))
    outputc = _getTimeStamp()+_FAIL+' ERROR: '+_getCallerName(frame)+(' '.join(args))+_ENDC
    print(outputc)
    _log(output + '\n')
    sys.exit(-1)


def TBRaise(*args,frame=2,exception=None):
    """
    Add custom text to exceptions, which can be useful for debugging sometimes. 

    Args:
        frame (int, optional): Controls the name of the caller. Defaults to method that called TBRaise.
        exception (Exception, optional): If None, raise generic ToolboxException. Otherwise raise exception.
    """
    args    = [str(s) for s in args]
    output  = _getCallerName(frame)+(' '.join(args))
    outputc = _WARNING+_getCallerName(frame)+(' '.join(args))+_ENDC
    _log(output + '\n')
    if exception is None:
        raise ToolboxException(output) 
    else:
        print(outputc)
        raise exception 


def TBPass(*args):
    args    = [str(s) for s in args]
    output  = _getTimeStamp()+' SUCCESS: '+(' '.join(args))
    outputc = _getTimeStamp()+_PASS+' SUCCESS: '+(' '.join(args))+_ENDC
    print(outputc)
    _log(output + '\n')

